"""Database abstraction layer for DuckDB

See base.py for design rules of thumb.

Since duckdb tends to be more malleable than a cloud-based database, if you
need to paper over differences, it's best to use the connection.create_function
pattern here to get sql flavors in alignment.
"""

import collections
import datetime
import pathlib
import re

import duckdb
import pandas
import pyarrow.dataset

from cumulus_library.databases import base


class DuckDatabaseBackend(base.DatabaseBackend):
    """Database backend that uses local files via duckdb"""

    def __init__(self, db_file: str, schema_name: str | None = None):
        super().__init__("main")
        self.db_type = "duckdb"
        self.db_file = db_file
        self.connection = None

    def init_errors(self):
        return ["Binder Error", "Catalog Error"]

    def connect(self):
        """Connects to the local duckdb database"""
        # As of the 1.0 duckdb release, local scopes, where schema names can be provided
        # as configuration to duckdb.connect, are not supported.
        # https://duckdb.org/docs/sql/statements/set.html#syntax
        # This is where the connection config would be supplied when it is supported
        self.connection = duckdb.connect(self.db_file)
        # Aliasing Athena's as_pandas to duckDB's df cast
        duckdb.DuckDBPyConnection.as_pandas = duckdb.DuckDBPyConnection.df

        # Ignore order of NDJSON that we load in. It saves memory and order doesn't matter for us.
        self.connection.execute("SET preserve_insertion_order = false;")

        # Paper over some syntax differences between Athena and DuckDB
        self.connection.create_function(
            # DuckDB's version is array_to_string -- seems there is no standard here.
            "array_join",
            self._compat_array_join,
            None,
            duckdb.typing.VARCHAR,
        )
        self.connection.create_function(
            # DuckDB's version is regexp_matches.
            "regexp_like",
            self._compat_regexp_like,
            None,
            duckdb.typing.BOOLEAN,
        )
        self.connection.create_function(
            # We frequently use Athena's date() function because it's easier than
            # the more widely-supported way of CAST(x AS DATE).
            # Rather than convert all of our SQL to the longer version,
            # we'll just make our own version of date().
            "date",
            self._compat_date,
            None,
            duckdb.typing.DATE,
        )
        self.connection.create_function(
            "from_iso8601_timestamp",
            self._compat_from_iso8601_timestamp,
            None,
            duckdb.typing.TIMESTAMP,
        )
        self.connection.create_function(
            # When trying to calculate an MD5 hash in Trino/Athena, the implementation
            # expects to recieve a varbinary type, so if you're hashing a string,
            # you would invoke it like `SELECT md5(to_utf8(string_col)) from table`.
            #
            # DuckDB's md5() function accepts a varchar instead, and does not have a
            # to_utf() function or varbinary type, so we patch this with a UDF that
            # just provides back the original string. As a result, these functions
            # have different signatures, but for cases like this where you're
            # conforming an argument to another function, it provides appropriate
            # function mocking
            #
            # NOTE: currently we do not have a use case beyond experimentation where
            # using MD5 hashes provide a benefit. Until we do, it is not required to
            # support this in other DatabaseBackend implementations.
            "to_utf8",
            self._compat_to_utf8,
            None,
            duckdb.typing.VARCHAR,
        )

    def insert_tables(self, tables: dict[str, pyarrow.dataset.Dataset]) -> None:
        """Ingests all ndjson data from a folder tree.

        This is often the output folder of Cumulus ETL"""
        for name, table in tables.items():
            self.connection.register(name, table)

    @staticmethod
    def _compat_array_join(value: list[str | None], delimiter: str | None) -> str:
        if delimiter is None or delimiter == "None":
            # This is exercised locally on unit tests but is not in CI. Not sure why,
            # and not sure it's worth debugging
            delimiter = ""  # pragma: no cover
        return delimiter.join(v for v in value if v is not None)

    @staticmethod
    def _compat_regexp_like(string: str | None, pattern: str | None) -> bool:
        match = re.search(pattern, string)
        return match is not None

    @staticmethod
    def _compat_date(
        value: str | datetime.datetime | datetime.date,
    ) -> datetime.date:
        if isinstance(value, str):
            return datetime.date.fromisoformat(value)
        elif isinstance(value, datetime.datetime):
            return value.date()
        elif isinstance(value, datetime.date):
            return value
        else:
            raise ValueError("Unexpected date() argument:", type(value), value)

    @staticmethod
    def _compat_to_utf8(value: str | None) -> datetime.date | None:
        """See the create_function() call for to_utf8 for more background"""
        # This is exercised locally on unit tests but is not in CI. Not sure why,
        # and not sure it's worth debugging
        return value  # pragma: no cover

    @staticmethod
    def _compat_from_iso8601_timestamp(
        value: str,
    ) -> datetime.datetime:
        # handle partial dates like 1970 or 1980-12 (which spec allows)
        if len(value) < 10:
            pieces = value.split("-")
            if len(pieces) == 1:
                return datetime.datetime(int(pieces[0]), 1, 1)
            else:
                return datetime.datetime(int(pieces[0]), int(pieces[1]), 1)

        # TODO: return timezone-aware datetimes, like Athena does
        #       (this currently generates naive datetimes, in UTC local time)
        return datetime.datetime.fromisoformat(value)

    def cursor(self) -> duckdb.DuckDBPyConnection:
        # Don't actually create a new connection,
        # because then we'd have to re-register our json tables.
        return self.connection

    def pandas_cursor(self) -> duckdb.DuckDBPyConnection:
        # Since this is not provided, return the vanilla cursor
        return self.connection

    def execute_as_pandas(
        self, sql: str, chunksize: int | None = None
    ) -> (pandas.DataFrame | collections.abc.Iterator[pandas.DataFrame], list[tuple]):
        # We call convert_dtypes here in case there are integer columns.
        # Pandas will normally cast nullable-int as a float type unless
        # we call this to convert to its nullable int column type.
        # PyAthena seems to do this correctly for us, but not DuckDB.
        result = self.connection.execute(sql)
        if chunksize:
            return iter([result.df().convert_dtypes()]), result.description
        return result.df().convert_dtypes(), result.description

    def parser(self) -> base.DatabaseParser:
        return DuckDbParser()

    def operational_errors(self) -> tuple[type[Exception], ...]:
        return (
            duckdb.OperationalError,
            duckdb.BinderException,
        )

    def export_table_as_parquet(
        self, table_name: str, file_name: str, location: pathlib.Path, *args, **kwargs
    ) -> bool:
        parquet_path = location / f"{file_name}"
        table_size = self.connection.execute(f"SELECT count(*) FROM {table_name}").fetchone()  # noqa: S608
        if table_size[0] == 0:
            return False
        query = f"""COPY
            (SELECT * FROM {table_name})
            TO '{parquet_path}'
            (FORMAT parquet)
            """  # noqa: S608
        self.connection.execute(query)
        return True

    def create_schema(self, schema_name):
        """Creates a new schema object inside the database"""
        schemas = self.connection.sql(
            "SELECT schema_name FROM information_schema.schemata"
        ).fetchall()
        if (schema_name,) not in schemas:
            self.connection.sql(f"CREATE SCHEMA {schema_name}")

    def close(self) -> None:
        if self.connection is not None:
            self.connection.close()


class DuckDbParser(base.DatabaseParser):
    """Table parser for DuckDB schemas"""

    def parse_found_schema(self, schema: dict[str, str]) -> dict:
        """parses a information_schema.tables query response for valid columns"""
        parsed = {}

        for key, value in schema.items():
            if isinstance(value, str):
                # DuckDB provides a parser to go from string -> type objects
                value = duckdb.typing.DuckDBPyType(value)

            # Collapse lists to the contained value
            if value.id == "list":
                value = value.children[0][1]  # [('child', CONTAINED_TYPE)]

            if value.id == "struct":
                result = self.parse_found_schema(dict(value.children))
            else:
                result = {}
            parsed[key.lower()] = result

        return parsed



"""Collection of jinja template getters for common SQL queries"""

from pathlib import Path

from cumulus_library.errors import CumulusLibraryError
from cumulus_library.template_sql import base_templates


def get_distinct_ids(
    columns: list[str],
    source_table: str,
    join_id: str | None = None,
    filter_table: str | None = None,
) -> str:
    """Gets distinct ids from a table, optionally filtering by ids in another table

    This is expected to be used in two ways:
    - To retrieve ids from a study cohort table
    - To retreive ids from a core FHIR table, excluding ids in a cohort study table

    It is expected that, if supplying optional parameters, all of them must be set

    :param columns: a list of ids to request
    :param source_table: the table to retrieve ids from
    :param join_id: the id column to use for joining. Expected to exist in both
      source and filter tables
    :param filter_table: a table containing ids you want to exclude
    """
    if (join_id and not filter_table) or (not join_id and filter_table):
        raise CumulusLibraryError(
            "psm_templates.get_distinct_ids expects all optional parameters ",
            "to be defined if supplied",
        )

    return base_templates.get_base_template(
        "psm_distinct_ids",
        Path(__file__).parent,
        columns=columns,
        source_table=source_table,
        join_id=join_id,
        filter_table=filter_table,
    )


def get_create_covariate_table(
    target_table: str,
    pos_source_table: str,
    neg_source_table: str,
    table_suffix: str,
    primary_ref: str,
    dependent_variable: str,
    join_cols_by_table: dict,
    count_ref: str | None = None,
    count_table: str | None = None,
) -> str:
    """Gets a query to create a covariate table for PSM analysis

    :param target_table: the name of the table to create
    :param pos_source_table: the table defining your positive cohort
    :param neg_source_table: the table you defined your positive cohort against,
        containing the full population
    :primary_ref: the ID field to use for IDing members of your cohort
    :dependent_variable: the name for the condition you are investigating, which
        will differentiate your positive and negative sources
    :join_cols_by_table: A dict defining extra data to join. See the psm_config.toml
        in tests/test_data/psm for more details
    :count_ref: optional ID to count records with for validation
    :count_table: optional table to use as the source of the count_refs
    """
    if (count_ref and not count_table) or (not count_ref and count_table):
        raise CumulusLibraryError(
            "psm_templates.get_create_covariate_table expects all count parameters ",
            "to be defined if supplied",
        )

    return base_templates.get_base_template(
        "psm_create_covariate_table",
        Path(__file__).parent,
        target_table=target_table,
        pos_source_table=pos_source_table,
        neg_source_table=neg_source_table,
        table_suffix=table_suffix,
        primary_ref=primary_ref,
        dependent_variable=dependent_variable,
        count_ref=count_ref,
        count_table=count_table,
        join_cols_by_table=join_cols_by_table,
    )



"""Package metadata"""

from cumulus_library.base_utils import StudyConfig
from cumulus_library.builders.base_table_builder import BaseTableBuilder
from cumulus_library.builders.counts import CountsBuilder
from cumulus_library.builders.statistics_templates.counts_templates import CountAnnotation
from cumulus_library.study_manifest import StudyManifest
from cumulus_library.template_sql.base_templates import get_template

# A note about the `get_template` function:
# As of this writing, the get_template function is used internally by the
# Cumulus team across multiple study repos for creating SQL in databases, both
# from the templates in this repo and in the various other projects (while leveraging
# some of our syntax helper macros, which are auto loaded by this function).
# Breaking changes to these templates will result in a major version bump.

# This API should be usable for your own study efforts - but the documentation
# is all code level. See template_sql for more information.

__all__ = [
    "BaseTableBuilder",
    "CountAnnotation",
    "CountsBuilder",
    "StudyConfig",
    "StudyManifest",
    "get_template",
]
__version__ = "4.6.2"



"""Manages configuration for argparse"""

import argparse

# Functions for arguments used by more than one sub-command


def add_aws_config(parser: argparse.ArgumentParser) -> None:
    """Adds arguments related to aws credentials to a subparser"""
    aws = parser.add_argument_group("AWS config")
    aws.add_argument("--profile", help="AWS profile", default="default")
    aws.add_argument(
        "--workgroup",
        default="cumulus",
        dest="work_group",
        help="Cumulus Athena workgroup (default: cumulus)",
    )
    aws.add_argument(
        "--region",
        help="AWS region data of Athena instance (default: us-east-1)",
        default="us-east-1",
    )


def add_custom_option(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-o",
        "--option",
        action="append",
        dest="options",
        help=(
            'Add a study-specific option, as "arg:value". '
            "See individual study documentation for possible argument names"
        ),
    )


def add_data_path_argument(parser: argparse.ArgumentParser) -> None:
    """Adds path arg to a subparser"""
    parser.add_argument(
        "data_path",
        default="./",
        nargs="?",
        help=(
            "The path to use for exporting counts data. "
            "Can be provided via CUMULUS_LIBRARY_DATA_PATH environment variable."
        ),
    )


def add_db_config(parser: argparse.ArgumentParser, input_mode: bool = False) -> None:
    """
    Adds arguments related to database backends to a subparser

    Pass input_mode=True if the subparser is for a command that operates on input
    (e.g. a build operation) rather than output tables (e.g. a clean operation)
    """
    group = parser.add_argument_group("Database config")
    group.add_argument(
        "--db-type",
        help="Which database backend to use (default athena)",
        choices=["athena", "duckdb"],
        default="athena",
    )
    group.add_argument(
        "--database",
        # In Athena, we use this as the schema_name (which is also called a Database
        # in their UX).
        #
        # In DuckDB, we use this as the path to the filename to store tables.
        #
        help="Database/schema name (for Athena) or file (for DuckDB)",
    )
    group.add_argument(
        "--schema_name",
        # In Athena, we use this as the schema_name.
        #
        # In DuckDB, we will eventually use this to allow for schema specification,
        # though it's in the bin of 'not implemented yet' for the way the python
        # API accesses it on connect.
        #
        help="Database/schema name (for Athena) or schema (for DuckDB)",
    )

    if input_mode:
        group.add_argument(
            "--load-ndjson-dir",
            help="Load ndjson files from this folder",
            metavar="DIR",
        )

    # Backend-specific config:
    add_aws_config(parser)


def add_study_dir_argument(parser: argparse.ArgumentParser) -> None:
    """Adds --study-dir arg to a subparser"""
    parser.add_argument(
        "-s",
        "--study-dir",
        action="append",
        help=(
            "Optionally add one or more directories to look for study definitions in. "
            "Default is in project directory and CUMULUS_LIBRARY_STUDY_DIR, "
            "if present, followed by any supplied paths. Target, and all its "
            "subdirectories, are checked for manifests. Overriding studies with the"
            " same namespace supersede earlier ones."
        ),
    )


def add_table_builder_argument(parser: argparse.ArgumentParser) -> None:
    """Adds --builder arg to a subparser"""
    parser.add_argument(
        "--builder",
        help=argparse.SUPPRESS,
    )


def add_target_argument(parser: argparse.ArgumentParser) -> None:
    """Adds --target arg to a subparser"""
    parser.add_argument(
        "-t",
        "--target",
        action="append",
        help=("Specify one or more studies to perform actions against."),
    )


def add_verbose_argument(parser: argparse.ArgumentParser) -> None:
    """Adds --verbose arg to a subparser"""
    parser.add_argument(
        "--verbose",
        default=False,
        action="store_true",
        help="Prints detailed SQL query info",
    )


# Parser construction


def get_parser() -> argparse.ArgumentParser:
    """Provides parser for handling CLI arguments"""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""Generates study tables and views from post-Cumulus ETL data.

cumulus-library will attempt to create a connection to a databse. If using 
AWS Athena, the following order of preference is used to select credentials:
  - explicit command line arguments
  - cumulus environment variables (see docs for more info)
  - Normal boto profile order (AWS env vars, ~/.aws/credentials, ~/.aws/config)""",
        epilog="See 'cumulus-library [action] -h' for usage of a specific action",
    )

    actions = parser.add_subparsers(
        title="actions",
        help="Available library actions",
        dest="action",
    )

    # Database cleaning

    clean = actions.add_parser(
        "clean", help="Removes tables & views beginning with '[target]__' from Athena"
    )

    add_custom_option(clean)
    add_db_config(clean)
    add_target_argument(clean)
    add_study_dir_argument(clean)
    add_verbose_argument(clean)

    clean.add_argument(
        "--prefix",
        action="store_true",
        help=argparse.SUPPRESS,
    )
    clean.add_argument(
        "--statistics",
        action="store_true",
        help="Remove artifacts of previous statistics runs",
        dest="stats_clean",
    )

    # Database building

    build = actions.add_parser(
        "build",
        help="Removes and recreates Athena tables & views for specified studies",
    )
    add_custom_option(build)
    add_data_path_argument(build)
    add_db_config(build, input_mode=True)
    add_study_dir_argument(build)
    add_table_builder_argument(build)
    add_target_argument(build)
    add_verbose_argument(build)

    build.add_argument(
        "--continue",
        dest="continue_from",
        help=argparse.SUPPRESS,
    )
    build.add_argument(
        "--force-upload",
        action="store_true",
        help="Forces file downloads/uploads to occur, even if they already exist",
    )
    build.add_argument(
        "--prepare",
        action="store_true",
        help=argparse.SUPPRESS,
    )
    build.add_argument(
        "--statistics",
        action="store_true",
        help=(
            "Force regenerating statistics data from latest dataset. "
            "Stats are created by default when study is initially run"
        ),
        dest="stats_build",
    )
    build.add_argument(
        "--umls-key",
        help="An API Key for the UMLS API",
    )

    # Database export

    export = actions.add_parser("export", help="Generates files on disk from Athena tables/views")
    add_custom_option(export)
    add_target_argument(export)
    add_study_dir_argument(export)
    add_data_path_argument(export)
    add_verbose_argument(export)
    add_db_config(export)
    export.add_argument(
        "--archive",
        action="store_true",
        help="Generates archive of :all: study tables, ignoring manifest export list.",
    )

    # Database import

    importer = actions.add_parser("import", help="Recreates a study from an exported archive")
    add_db_config(importer)
    add_verbose_argument(importer)
    importer.add_argument(
        "-a",
        "--archive-path",
        action="append",
        help="The path to an archive generated by the export CLI subcommand",
    )
    # Aggregator upload

    upload = actions.add_parser("upload", help="Bulk uploads data to Cumulus aggregator")
    add_data_path_argument(upload)
    add_target_argument(upload)

    upload.add_argument("--id", help="Site ID. Default is value of CUMULUS_AGGREGATOR_ID")
    upload.add_argument(
        "--preview",
        default=False,
        action="store_true",
        help="Run pre-fetch and prepare upload, but log output instead of sending.",
    )
    upload.add_argument(
        "--url",
        help=(
            "Upload URL. Default is value of CUMULUS_AGGREGATOR_URL if present, "
            "or smart cumulus instance"
        ),
        default="https://aggregator.smartcumulus.org/",
    )
    upload.add_argument(
        "--network",
        help=(
            "Network name. Reach out to your aggregator administrator for the appropriate value."
        ),
    )
    upload.add_argument("--user", help="Cumulus user. Default is value of CUMULUS_AGGREGATOR_USER")

    # Generate a study's template-driven sql

    sql = actions.add_parser(
        "generate-sql", help="Generates a study's template-driven sql for reference"
    )
    add_custom_option(sql)
    add_db_config(sql, input_mode=True)
    add_table_builder_argument(sql)
    add_target_argument(sql)
    add_study_dir_argument(sql)

    # Generate markdown tables for documentation

    markdown = actions.add_parser(
        "generate-md", help="Generates markdown tables for study documentation"
    )
    add_data_path_argument(markdown)
    add_db_config(markdown)
    add_study_dir_argument(markdown)
    add_target_argument(markdown)
    add_verbose_argument(markdown)

    # Get study version

    version = actions.add_parser("version", help="Gets the versions of the CLI and studies")
    add_study_dir_argument(version)

    return parser



#!/usr/bin/env python3
"""Utility for building/retrieving data views in AWS Athena"""

import copy
import importlib.util
import json
import os
import pathlib
import sys

import requests
import rich

from cumulus_library import (
    __version__,
    base_utils,
    cli_parser,
    databases,
    enums,
    errors,
    log_utils,
    study_manifest,
)
from cumulus_library.actions import (
    builder,
    cleaner,
    exporter,
    file_generator,
    importer,
    uploader,
)


class StudyRunner:
    """Class for managing cursors and invoking actions"""

    verbose = False
    schema_name = None

    def __init__(self, config: base_utils.StudyConfig, data_path: str | None):
        self.config = config
        self.data_path = data_path and pathlib.Path(data_path)

    def get_config(self, manifest: study_manifest.StudyManifest):
        schema = base_utils.get_schema(self.config, manifest)
        if schema == self.config.schema:
            return self.config
        else:
            config = copy.copy(self.config)
            config.schema = schema
            return config

    ### Creating studies

    def clean_study(
        self,
        targets: list[str],
        study_dict: dict,
        *,
        options: dict[str, str],
        prefix: bool = False,
    ) -> None:
        """Removes study table/views from Athena.

        While this is usually not required, since it it done as part of a build,
        this can be useful for cleaning up tables if a study prefix is changed
        for some reason.

        :param targets: The study prefixes to use for IDing tables to remove
        :param study_dict: The dictionary of available study targets
        :param options: The dictionary of study-specific options
        :keyword prefix: If True, does a search by string prefix in place of study name
        """
        for target in targets:
            if prefix:
                manifest = study_manifest.StudyManifest(options=options)
                cleaner.clean_study(
                    config=self.get_config(manifest), manifest=manifest, prefix=target
                )
            else:
                manifest = study_manifest.StudyManifest(study_dict[target], options=options)
                cleaner.clean_study(config=self.get_config(manifest), manifest=manifest)

    def clean_and_build_study(
        self,
        target: pathlib.Path,
        *,
        options: dict[str, str],
        continue_from: str | None = None,
        prepare: bool = False,
        data_path: pathlib.Path | None = None,
    ) -> None:
        """Recreates study views/tables

        :param target: A path to the study directory
        :keyword options: The dictionary of study-specific options
        :keyword continue_from: Restart a run from a specific sql file (for dev only)
        :keyword prepare: If true, will render query instead of executing
        :keyword data_path: If prepare is true, the path to write rendered data to
        """
        manifest = study_manifest.StudyManifest(target, self.data_path, options=options)
        try:
            if not prepare:
                builder.run_protected_table_builder(
                    config=self.get_config(manifest), manifest=manifest
                )
                if not continue_from:
                    log_utils.log_transaction(
                        config=self.get_config(manifest),
                        manifest=manifest,
                        status=enums.LogStatuses.STARTED,
                    )
                    cleaner.clean_study(
                        config=self.get_config(manifest),
                        manifest=manifest,
                    )
                else:
                    log_utils.log_transaction(
                        config=self.get_config(manifest),
                        manifest=manifest,
                        status=enums.LogStatuses.RESUMED,
                    )

            builder.build_study(
                config=self.get_config(manifest),
                manifest=manifest,
                continue_from=continue_from,
                data_path=data_path,
                prepare=prepare,
            )
            if not prepare:
                log_utils.log_transaction(
                    config=self.get_config(manifest),
                    manifest=manifest,
                    status=enums.LogStatuses.FINISHED,
                )

        except errors.StudyManifestFilesystemError as e:
            # This should be thrown prior to any database connections, so
            # skipping logging
            raise e  # pragma: no cover
        except Exception as e:
            if not prepare:
                log_utils.log_transaction(
                    config=self.get_config(manifest),
                    manifest=manifest,
                    status=enums.LogStatuses.ERROR,
                )
            raise e

    def build_matching_files(
        self,
        target: pathlib.Path,
        table_builder_name: str,
        *,
        options: dict[str, str],
        prepare: bool = False,
        data_path: pathlib.Path,
    ) -> None:
        """Runs a single table builder

        :param target: A path to the study directory
        :param table_builder_name: a builder file referenced in the study's manifest
        :keyword options: The dictionary of study-specific options
        :keyword prepare: If true, will render query instead of executing
        :keyword data_path: If prepare is true, the path to write rendered data to
        """
        manifest = study_manifest.StudyManifest(target, options=options)
        builder.build_matching_files(
            config=self.get_config(manifest),
            manifest=manifest,
            builder=table_builder_name,
            prepare=prepare,
            data_path=data_path,
        )

    ### Data exporters
    def export_study(
        self,
        target: pathlib.Path,
        data_path: pathlib.Path,
        archive: bool,
        *,
        options: dict[str, str],
    ) -> None:
        """Exports aggregates defined in a manifest

        :param target: A path to the study directory
        :param data_path: The location to export data to
        :param archive: If true, will export all tables, otherwise uses manifest list
        :param options: The dictionary of study-specific options
        """
        manifest = study_manifest.StudyManifest(target, data_path, options=options)
        exporter.export_study(
            config=self.get_config(manifest),
            manifest=manifest,
            data_path=data_path,
            archive=archive,
        )

    def generate_study_sql(
        self,
        target: pathlib.Path,
        *,
        options: dict[str, str],
    ) -> None:
        """Materializes study sql from templates

        :param target: A path to the study directory
        :param options: The dictionary of study-specific options
        """
        manifest = study_manifest.StudyManifest(target, options=options)
        file_generator.run_generate_sql(config=self.get_config(manifest), manifest=manifest)

    def generate_study_markdown(
        self,
        target: pathlib.Path,
    ) -> None:
        """Materializes study sql from templates

        :param target: A path to the study directory
        """
        manifest = study_manifest.StudyManifest(target)
        file_generator.run_generate_markdown(config=self.get_config(manifest), manifest=manifest)


def get_abs_path(path: str) -> pathlib.Path:
    """Convenience method for handling abs vs rel paths"""
    if path[0] == "/":
        return pathlib.Path(path)
    else:
        return pathlib.Path(pathlib.Path.cwd(), path)


def get_study_dict(alt_dir_paths: list) -> dict[str, pathlib.Path] | None:
    """Gets valid study targets from ./studies/, and any pip installed studies

    :returns: A list of Path objects
    """
    manifest_studies = {}
    cli_path = pathlib.Path(__file__).resolve().parents[0]

    # first, we'll take note of our own internal studies
    paths = [pathlib.Path(cli_path, "studies")]

    # then, we'll get any installed public studies
    with open(
        pathlib.Path(cli_path, "./module_allowlist.json"), encoding="utf-8"
    ) as study_allowlist_json:
        study_allowlist = json.load(study_allowlist_json)["allowlist"]
    for module_name in study_allowlist:
        if spec := importlib.util.find_spec(module_name):
            paths += [pathlib.Path(x) for x in spec.submodule_search_locations]

    # lastly, any user supplied paths. These take precedence.
    if alt_dir_paths is not None:
        paths = paths + alt_dir_paths

    # Now actually search all the paths for manifest.toml files
    for path in paths:
        found_studies = get_studies_by_manifest_path(path)
        manifest_studies.update(found_studies)
    return manifest_studies


def get_studies_by_manifest_path(path: pathlib.Path) -> dict[str, pathlib.Path]:
    """Recursively search for manifest.toml files from a given path"""
    manifest_paths = {}
    for child_path in path.iterdir():
        if child_path.is_dir():
            manifest_paths.update(get_studies_by_manifest_path(child_path))
        elif child_path.name == "manifest.toml":
            try:
                # Pass empty options, since we are not in a study-specific context
                manifest = study_manifest.StudyManifest(path, options={})
                manifest_paths[manifest.get_study_prefix()] = path
            except errors.StudyManifestParsingError as exc:
                rich.print(f"[bold red] Ignoring study in '{path}': {exc}")
    return manifest_paths


def run_cli(args: dict):
    """Controls which library tasks are run based on CLI arguments"""
    console = rich.get_console()
    if args["action"] != "import" and not args.get("target"):
        sys.exit("Please specify one or more studies with `-t [study_name]`.")
    if args["action"] == "upload":
        try:
            uploader.upload_files(args)
        except requests.RequestException as e:  # pragma: no cover
            rich.print(str(e))
            sys.exit()

    # all other actions require connecting to the database
    else:
        # if we're targeting a builder explicitly, we always want to
        # run in drop mode
        if builder := args.get("builder"):
            drop_table = builder
        else:
            drop_table = args.get("drop_table")
        db, schema = databases.create_db_backend(args)
        config = base_utils.StudyConfig(
            db=db,
            schema=schema,
            drop_table=drop_table,
            force_upload=args.get("replace_existing", False),
            verbose=args.get("verbose"),
            stats_build=args.get("stats_build", False),
            stats_clean=args.get("stats_clean", False),
            umls_key=args.get("umls_key"),
            options=args.get("options"),
        )
        try:
            runner = StudyRunner(config, data_path=args.get("data_path"))
            if not args.get("prepare"):
                console.print("[italic] Connecting to database...")
                runner.config.db.cursor().execute("SHOW DATABASES")
            study_dict = get_study_dict(args.get("study_dir"))
            if "prefix" not in args.keys():
                if args.get("target"):
                    for target in args["target"]:
                        if target not in study_dict:
                            sys.exit(
                                f"{target} was not found in available studies: "
                                f"{list(study_dict.keys())}.\n\n"
                                "If you are trying to run a custom study, make sure "
                                "you include `-s path/to/study/dir` as an argument."
                            )
            if args["action"] == "clean":
                runner.clean_study(
                    targets=args["target"],
                    study_dict=study_dict,
                    prefix=args["prefix"],
                    options=args["options"],
                )
            elif args["action"] == "build":
                for target in args["target"]:
                    if args["builder"]:
                        runner.build_matching_files(
                            study_dict[target],
                            args["builder"],
                            options=args["options"],
                            prepare=args["prepare"],
                            data_path=args["data_path"],
                        )
                    else:
                        runner.clean_and_build_study(
                            study_dict[target],
                            continue_from=args["continue_from"],
                            options=args["options"],
                            prepare=args["prepare"],
                            data_path=args["data_path"],
                        )

            elif args["action"] == "export":
                if args["archive"]:
                    warning_text = (
                        "🚨[bold red] This will export all study tables [/bold red]🚨"
                        "\n\nDepending on your study definition, this data may contain "
                        "data that would be characterized as a [italic]limited data "
                        "set[/italic], primarily dates, on a per patient level.\n\n"
                        "[bold]By doing this, you are assuming the responsibility for "
                        "meeting your organization's security requirements for "
                        "storing this data in a secure manner.[/bold]\n\n"
                        "Type Y to proceed, or any other value to quit.\n"
                    )
                    console.print(warning_text)
                    response = input()
                    if response.lower() != "y":
                        sys.exit()
                for target in args["target"]:
                    runner.export_study(
                        study_dict[target],
                        args["data_path"],
                        args["archive"],
                        options=args["options"],
                    )

            elif args["action"] == "import":
                for archive in args["archive_path"]:
                    archive = get_abs_path(archive)
                    importer.import_archive(config, archive_path=archive)

            elif args["action"] == "generate-sql":
                for target in args["target"]:
                    runner.generate_study_sql(study_dict[target], options=args["options"])

            elif args["action"] == "generate-md":
                for target in args["target"]:
                    runner.generate_study_markdown(study_dict[target])
        finally:
            config.db.close()
        # For unit testing only
        return config


def main(cli_args=None):
    """Reads CLI input/environment variables and invokes library calls"""

    parser = cli_parser.get_parser()
    args = vars(parser.parse_args(cli_args))
    console = rich.get_console()
    arg_env_pairs = (
        ("data_path", "CUMULUS_LIBRARY_DATA_PATH"),
        ("db_type", "CUMULUS_LIBRARY_DB_TYPE"),
        ("id", "CUMULUS_AGGREGATOR_ID"),
        ("load_ndjson_dir", "CUMULUS_LIBRARY_LOAD_NDJSON_DIR"),
        ("profile", "CUMULUS_LIBRARY_PROFILE"),
        ("region", "CUMULUS_LIBRARY_REGION"),
        ("schema_name", "CUMULUS_LIBRARY_SCHEMA_NAME"),
        ("database", "CUMULUS_LIBRARY_DATABASE"),
        ("study_dir", "CUMULUS_LIBRARY_STUDY_DIR"),
        ("umls_key", "UMLS_API_KEY"),
        ("url", "CUMULUS_AGGREGATOR_URL"),
        ("user", "CUMULUS_AGGREGATOR_USER"),
        ("network", "CUMULUS_AGGREGATOR_NETWORK"),
        ("work_group", "CUMULUS_LIBRARY_WORKGROUP"),
    )
    read_env_vars = []
    for pair in arg_env_pairs:
        if env_val := os.environ.get(pair[1]):
            if pair[0] == "study_dir":
                args[pair[0]] = [env_val]
            else:
                args[pair[0]] = env_val
            read_env_vars.append([pair[1], env_val])

    # We process this arg first, since version checking uses it
    if args.get("study_dir"):
        posix_paths = []
        for path in args["study_dir"]:
            posix_paths.append(get_abs_path(path))
        args["study_dir"] = posix_paths

    if args["action"] is None:
        sys.exit("No actions selected. Run 'cumulus-library -h' for details about actions.")

    if args["action"] == "version":
        table = rich.table.Table(title=f"cumulus-library version: {__version__}")
        table.add_column("Study Name", style="green")
        table.add_column("Version", style="blue")
        studies = get_study_dict(args.get("study_dir"))
        for study in sorted(studies.keys()):
            try:
                spec = importlib.util.spec_from_file_location(
                    "study_init", studies[study] / "__init__.py"
                )
                study_init = importlib.util.module_from_spec(spec)
                sys.modules["study_init"] = study_init
                spec.loader.exec_module(study_init)
                table.add_row(study, study_init.__version__)
            except Exception:
                table.add_row(study, "No version defined")
        console.print(table)
        sys.exit(0)

    if len(read_env_vars) > 0:
        table = rich.table.Table(title="Values read from environment variables")
        table.add_column("Environment Variable", style="green")
        table.add_column("Value", style="cyan")
        for row in read_env_vars:
            if row[0] == "CUMULUS_AGGREGATOR_ID":
                table.add_row(row[0], "#########")  # pragma: no cover
            else:
                table.add_row(row[0], row[1])

        console.print(table)

    options = {}
    cli_options = args.get("options") or []
    for c_arg in cli_options:
        c_arg = c_arg.split(":", 1)
        if len(c_arg) == 1:
            sys.exit(
                f"Custom argument '{c_arg}' is not validly formatted.\n"
                "Custom arguments should be of the form 'argname:value'."
            )
        options[c_arg[0]] = c_arg[1]
    args["options"] = options

    if args.get("data_path"):
        args["data_path"] = get_abs_path(args["data_path"])
    return run_cli(args)


def main_cli():  # called by the generated wrapper scripts
    main()  # pragma: no cover


if __name__ == "__main__":
    main()  # pragma: no cover



import pathlib
import sys
from collections.abc import Iterable

import cumulus_fhir_support
import pyarrow
import pyarrow.dataset
import pyarrow.json
import rich

from cumulus_library import base_utils, db_config, errors
from cumulus_library.databases import athena, base, duckdb


def _list_files_for_resource(path: pathlib.Path, resource: str) -> list[str]:
    files = []

    # Support any ndjson files from the target folder directly
    files += list(cumulus_fhir_support.list_multiline_json_in_dir(path, resource))

    # Also support being given an ETL output folder, and look in the table subdir
    subdir = path / resource.lower()
    files += list(cumulus_fhir_support.list_multiline_json_in_dir(subdir, resource))

    return files


def _rows_from_files(files: list[str]) -> Iterable[dict]:
    for file in files:
        yield from cumulus_fhir_support.read_multiline_json(file)


def _json_format():
    """Returns a Dataset format object suitable for reading FHIR NDJSON."""
    # FHIR can have very long JSON lines (think DocRefs with inlined notes). And unfortunately,
    # the way PyArrow's JSON parser works is that each entire line must fit inside a single read
    # block, or an exception is raised. I've seen 13MB JSON lines in the wild, at least.
    # So let's be generous here and allow 50MB lines. And cross our fingers that it's enough.
    read_options = pyarrow.json.ReadOptions(block_size=50 * 2**20)
    return pyarrow.dataset.JsonFileFormat(read_options=read_options)


def _load_custom_etl_table(path: str) -> pyarrow.dataset.Dataset | None:
    """Loads a non-FHIR ETL table from disk (tables like etl__completion)."""
    files = list(cumulus_fhir_support.list_multiline_json_in_dir(path))
    if not files:
        return None

    # This is a custom Cumulus ETL table, with no easy way to get a schema definition.
    # We **could** put a hard-coded shared schema in cumulus-fhir-support, but since these
    # tables are so simple, we mostly don't yet need to have that level of pre-coordination.
    #
    # Instead, we let PyArrow infer the types, with one hiccup: its JSON parser will interpret
    # datetime strings as a TIMESTAMP type, which isn't insane, but not what we want - we want to
    # keep those as strings. So we let PyArrow infer from Python objects, where it works like we
    # want.
    #
    # Thus, we sip on the data a bit to infer a schema from the first row as a Python object.
    first_row = next(iter(_rows_from_files(files)), None)
    if not first_row:
        # This path shouldn't happen, since list_multiline_json_in_dir() won't give us back
        # empty files. But just in case.
        return None  # pragma: no cover
    schema = pyarrow.Table.from_pylist([first_row]).schema

    # Now let PyArrow load the rest of the data from disk on demand, rather than loading it all
    # into memory, but using the detected schema rather than inferring it with timestamps.
    return pyarrow.dataset.dataset(files, schema=schema, format=_json_format())


def read_ndjson_dir(path: str) -> dict[str, pyarrow.dataset.Dataset]:
    """Loads a directory tree of raw ndjson into schema-ful tables.

    :param path: a directory path
    :returns: dictionary of table names (like 'documentreference') to table
      data (with schema)
    """
    all_tables = {}

    # Manually specify the list of resources because we want to create each table
    # even if the folder does not exist.
    resources = [
        "AllergyIntolerance",
        "Condition",
        "Device",
        "DiagnosticReport",
        "DocumentReference",
        "Encounter",
        "Immunization",
        "Medication",
        "MedicationRequest",
        "Observation",
        "Patient",
        "Procedure",
        "ServiceRequest",
    ]
    for resource in resources:
        table_name = resource.lower()
        files = _list_files_for_resource(pathlib.Path(path), resource)

        # Make a pyarrow table with full schema from the data
        schema = cumulus_fhir_support.pyarrow_schema_from_rows(resource, _rows_from_files(files))
        # Use a PyArrow Dataset (vs a Table) to avoid loading all the files in memory.
        all_tables[table_name] = pyarrow.dataset.dataset(
            files, schema=schema, format=_json_format()
        )

    # And now some special support for a few ETL tables.
    metadata_tables = [
        "etl__completion",
        "etl__completion_encounters",
    ]
    for metadata_table in metadata_tables:
        if dataset := _load_custom_etl_table(f"{path}/{metadata_table}"):
            all_tables[metadata_table] = dataset

    return all_tables


def _handle_load_ndjson_dir(args: dict[str, str], backend: base.DatabaseBackend) -> None:
    load_ndjson_dir = args.get("load_ndjson_dir")
    if not load_ndjson_dir:
        return  # nothing to do

    if db_config.db_type != "duckdb":
        sys.exit("Loading an NDJSON dir is only supported with --db-type=duckdb.")

    if backend.connection is None:
        return  # connect() was never run, we don't have a live DB connection

    with base_utils.get_progress_bar() as progress:
        progress.add_task("Detecting JSON schemas...", total=None)
        backend.insert_tables(read_ndjson_dir(load_ndjson_dir))


def create_db_backend(args: dict[str, str]) -> (base.DatabaseBackend, str):
    """Retrieves a database backend and target schema from CLI args"""
    db_config.db_type = args["db_type"]

    if db_config.db_type == "duckdb":
        # TODO: reevaluate as DuckDB's local schema support evolves.
        # https://duckdb.org/docs/sql/statements/set.html#syntax
        if not (args.get("schema_name") is None or args["schema_name"] == "main"):
            rich.print(  # pragma: no cover
                "Warning - local schema names are not yet supported by duckDB's "
                "python library - using 'main' instead"
            )
        schema_name = "main"
        backend = duckdb.DuckDatabaseBackend(args["database"])
    elif db_config.db_type == "athena":
        if (
            args.get("schema_name") is not None
            and args.get("database") is not None
            and args.get("schema_name") != args.get("database")
        ):
            sys.exit(
                f"Two separate values, database: '{args['database']}' and "
                f"schema_name: '{args['schema_name']}' were supplied for "
                "an Athena database connection.\n"
                "Only one of these needs to be supplied. Both can be passed "
                "as arguments, but they must be equal"
            )
        schema_name = args["schema_name"] or args["database"]
        backend = athena.AthenaDatabaseBackend(
            args["region"],
            args["work_group"],
            args["profile"],
            schema_name,
        )
    else:
        raise errors.CumulusLibraryError(f"'{db_config.db_type}' is not a supported database.")
    if "prepare" not in args.keys():
        backend.connect()
    elif not args["prepare"]:
        backend.connect()
    _handle_load_ndjson_dir(args, backend)
    return (backend, schema_name)



"""Collection of small commonly used utility functions"""

import dataclasses
import datetime
import pathlib
import shutil
import zipfile
from contextlib import contextmanager

import rich
from rich import progress

from cumulus_library import databases, study_manifest


@dataclasses.dataclass
class StudyConfig:
    """Class for containing study-centric parameters

    The intent of this class is that if you want something passed through to the
    prepare_queries section of a study, this is the place it should go. If you're
    doing something above that level, consider explicit arguments instead. This should
    be an interface aimed at a study author.

    :param db: a databaseBackend object for a specific target database
    :param schema: the database schema specified at the command line
    :keyword drop_table: when creating tables, if one already exists in the db,
        drop it
    :keyword force_upload: If the study downloads data from an external resource,
        force it to skip any cached data when running
    :keyword verbose: if True, print raw queries to the cli instead of progress bars
    :keyword stats_build: If the study runs a stats builder, force regeneration of
        any sampling or other stochastic techniques
    :keyword umls_key: An API for the UMLS service, used for downloading vocabularies
    :keyword umls: A UMLS API key
    :keyword options: a dictionary for any study-specific CLI arguments
    """

    db: databases.DatabaseBackend
    schema: str
    drop_table: bool = False
    force_upload: bool = False
    verbose: bool = False
    stats_build: bool = False
    stats_clean: bool = False
    umls_key: str | None = None
    options: dict | None = None


def get_schema(config: StudyConfig, manifest: study_manifest.StudyManifest):
    if dedicated := manifest.get_dedicated_schema():
        config.db.create_schema(dedicated)
        return dedicated
    return config.schema


def load_text(path: str) -> str:
    with open(path, encoding="UTF-8") as fp:
        return fp.read()


def parse_sql(sql_text: str) -> list[str]:
    commands = []

    for statement in sql_text.split(";"):
        parsed = []
        for line in statement.splitlines():
            if not line.strip().startswith("--"):
                parsed.append(line)
        commands.append("\n".join(parsed))
    return filter_strip(commands)


def filter_strip(commands) -> list[str]:
    return list(filter(None, [c.strip() for c in commands]))


@contextmanager
def query_console_output(
    verbose: bool, query: str, progress_bar: progress.Progress, task: progress.Task
):
    """Convenience context manager for handling console output"""
    if verbose:
        rich.print()
        rich.print(query)
    yield
    if not verbose:
        progress_bar.advance(task)


def get_progress_bar(**kwargs) -> progress.Progress:
    # The default columns don't change to elapsed time when finished.
    return progress.Progress(
        progress.TextColumn("[progress.description]{task.description}"),
        progress.BarColumn(),
        progress.TaskProgressColumn(),
        progress.TimeElapsedColumn(),
        **kwargs,
    )


def get_utc_datetime() -> datetime.datetime:
    return datetime.datetime.now(datetime.UTC).replace(microsecond=0)


def get_tablename_safe_iso_timestamp() -> str:
    """formats a timestamp to remove sql unallowed characters in table names"""
    iso_timestamp = get_utc_datetime().isoformat()
    safe_timestamp = iso_timestamp.replace(":", "_").replace("-", "_").replace("+", "_")
    return safe_timestamp


def zip_dir(read_path, write_path, archive_name):
    """Moves a directory to an archive"""
    file_list = [file for file in read_path.glob("**/*") if file.is_file()]
    timestamp = get_utc_datetime().isoformat().replace("+00:00", "Z")
    with zipfile.ZipFile(
        f"{write_path}/{archive_name}_{timestamp}.zip",
        "w",
        zipfile.ZIP_DEFLATED,
    ) as f:
        for file in file_list:
            f.write(file, file.relative_to(read_path))
            file.unlink()
        shutil.rmtree(read_path)


def update_query_if_schema_specified(query: str, manifest: study_manifest.StudyManifest):
    if manifest and manifest.get_dedicated_schema():
        # External queries in athena require a schema to be specified already, so
        # rather than splitting and ending up with a table name like
        # `schema`.`schema.table`, we just remove the `schema__` chunk from the
        # table name.
        # TODO: Move this to be a database defined function.
        if "CREATE EXTERNAL" in query:
            query = query.replace(
                f"{manifest.get_study_prefix()}__",
                "",
            )
        else:
            query = query.replace(
                f"{manifest.get_study_prefix()}__",
                f"{manifest.get_dedicated_schema()}.",
            )
    return query


def unzip_file(file_path: pathlib.Path, write_path: pathlib.Path):
    """Expands a zip archive"""
    with zipfile.ZipFile(file_path, mode="r") as z:
        for file in z.namelist():
            z.extract(file, write_path)



from pathlib import Path

import cumulus_library


class CoreCountsBuilder(cumulus_library.CountsBuilder):
    display_text = "Creating core counts..."

    def count_core_allergyintolerance(self, duration: str = "month"):
        table_name = self.get_table_name("count_allergyintolerance", duration=duration)
        from_table = self.get_table_name("allergyintolerance")
        cols = [
            ["category", "varchar", None],
            [f"recordedDate_{duration}", "date", None],
            ["code_display", "varchar", None],
            ["reaction_manifestation_display", "varchar", None],
        ]
        return self.count_allergyintolerance(table_name, from_table, cols)

    def count_core_condition(self, duration: str = "month"):
        table_name = self.get_table_name("count_condition", duration=duration)
        from_table = self.get_table_name("condition")
        cols = [
            ["category_code", "varchar", None],
            [f"recordedDate_{duration}", "date", None],
            ["code_display", "varchar", None],
            ["code", "varchar", None],
        ]
        return self.count_condition(table_name, from_table, cols)

    def count_core_diagnosticreport(self, duration: str = "month"):
        table_name = self.get_table_name("count_diagnosticreport", duration=duration)
        from_table = self.get_table_name("diagnosticreport")
        cols = [
            ["category_display", "varchar", None],
            ["code_display", "varchar", None],
            # Issued is not the _preferred_ time to pull, since it is an administrative time,
            # not a clinical one. But the clinical dates are annoyingly spread across three
            # fields: effectiveDateTime, effectivePeriod.start, and effectivePeriod.end.
            # So rather than do some fancy collation, just use issued. These core counts are
            # just a rough idea of the data, not a polished final product.
            [f"issued_{duration}", "date", None],
        ]
        return self.count_diagnosticreport(table_name, from_table, cols)

    def count_core_documentreference(self, duration: str = "month"):
        table_name = self.get_table_name("count_documentreference", duration=duration)
        from_table = self.get_table_name("documentreference")
        cols = [
            ["type_display", "varchar", None],
            [f"author_{duration}", "date", None],
        ]
        return self.count_documentreference(table_name, from_table, cols)

    def count_core_encounter(self, duration: str | None = None):
        table_name = self.get_table_name("count_encounter", duration=duration)
        from_table = self.get_table_name("encounter")

        cols = [
            f"period_start_{duration}",
            "class_display",
            "age_at_visit",
            "gender",
            "race_display",
            "ethnicity_display",
        ]

        return self.count_encounter(table_name, from_table, cols)

    def _count_core_encounter_type(self, table_name: str, cols: list, duration: str | None = None):
        """
        Encounter Type information is for every visit, and therefore this
        SQL should be precise in which fields to select (This is a BIG query).

        :param table_name: name of the view from "core__encounter_type"
        :param cols: from "core__encounter_type"
        :param duration: None or ''month', 'year'
        :return: A SQL statement as a string
        """
        table_name = self.get_table_name(table_name, duration)
        from_table = self.get_table_name("encounter")

        if duration:
            cols.append(f"period_start_{duration}")

        return self.count_encounter(table_name, from_table, cols)

    def count_core_encounter_all_types(self, duration: str | None = None):
        cols = [
            "class_display",
            "type_display",
            "serviceType_display",
            "priority_display",
        ]
        return self._count_core_encounter_type("count_encounter_all_types", cols, duration)

    # The following encounter tables all count on one specific code type

    def count_core_encounter_enc_type(self, duration: str = "month"):
        cols = ["class_display", "type_display"]
        return self._count_core_encounter_type("count_encounter_type", cols, duration)

    def count_core_encounter_priority(self, duration: str = "month"):
        cols = ["class_display", "priority_display"]
        return self._count_core_encounter_type("count_encounter_priority", cols, duration)

    def count_core_encounter_service(self, duration: str = "month"):
        cols = ["class_display", "serviceType_display"]
        return self._count_core_encounter_type("count_encounter_service", cols, duration)

    def count_core_medicationrequest(self, duration: str = "month"):
        table_name = self.get_table_name("count_medicationrequest", duration=duration)
        from_table = self.get_table_name("medicationrequest")
        cols = ["status", "intent", f"authoredon_{duration}", "medication_display"]
        return self.count_medicationrequest(table_name, from_table, cols)

    def count_core_observation_lab(self, duration: str = "month"):
        table_name = self.get_table_name("count_observation_lab", duration=duration)
        from_table = self.get_table_name("observation_lab")
        cols = [
            f"effectiveDateTime_{duration}",
            "observation_code",
            "valueCodeableConcept_display",
        ]
        return self.count_observation(table_name, from_table, cols)

    def count_core_patient(self):
        table_name = self.get_table_name("count_patient")
        from_table = self.get_table_name("patient")
        cols = ["gender", "race_display", "ethnicity_display"]
        return self.count_patient(table_name, from_table, cols)

    def count_core_procedure(self, duration: str = "month"):
        table_name = self.get_table_name("count_procedure", duration=duration)
        from_table = self.get_table_name("procedure")
        cols = [
            ["category_display", "varchar", None],
            ["code_display", "varchar", None],
            # The performed date is annoyingly spread across three fields: performedDateTime,
            # performedPeriod.start, and performedPeriod.end.
            # Rather than do some fancy collation, we just use performedDateTime.
            # It's the only "must support" performed field, and period seems less common.
            # These core counts are just a rough idea of the data, not a polished final product.
            [f"performedDateTime_{duration}", "date", None],
        ]
        return self.count_procedure(table_name, from_table, cols)

    def prepare_queries(self, *args, **kwargs):
        super().prepare_queries(*args, **kwargs)
        self.queries = [
            self.count_core_allergyintolerance(duration="month"),
            self.count_core_condition(duration="month"),
            self.count_core_diagnosticreport(duration="month"),
            self.count_core_documentreference(duration="month"),
            self.count_core_encounter(duration="month"),
            self.count_core_encounter_all_types(),
            self.count_core_encounter_all_types(duration="month"),
            self.count_core_encounter_enc_type(duration="month"),
            self.count_core_encounter_service(duration="month"),
            self.count_core_encounter_priority(duration="month"),
            self.count_core_medicationrequest(duration="month"),
            self.count_core_observation_lab(duration="month"),
            self.count_core_procedure(duration="month"),
            self.count_core_patient(),
        ]


if __name__ == "__main__":  # pragma: no cover
    builder = CoreCountsBuilder()
    builder.write_counts(f"{Path(__file__).resolve().parent}/count_core.sql")



from dataclasses import dataclass

import cumulus_library
from cumulus_library.studies.core.core_templates import core_templates
from cumulus_library.template_sql import sql_utils

expected_table_cols = {
    "encounter": {
        "id": [],
        "status": [],
        "period": [
            "start",
            "end",
        ],
        "class": sql_utils.CODING,
        "subject": sql_utils.REFERENCE,
    },
    "etl__completion": {
        "group_name": [],
    },
    "etl__completion_encounters": {
        "group_name": [],
    },
}


@dataclass(kw_only=True)
class EncConfig(sql_utils.CodeableConceptConfig):
    """Convenience wrapper for CodeableConceptConfig"""

    source_table: str = "encounter"

    def __post_init__(self):
        self.target_table = f"core__encounter_dn_{self.column_hierarchy[-1][0]}"


class CoreEncounterBuilder(cumulus_library.BaseTableBuilder):
    display_text = "Creating Encounter tables..."

    def denormalize_codes(self, database):
        code_configs = [
            EncConfig(
                column_hierarchy=[("type", list)],
                filter_priority=True,
                code_systems=[
                    "http://terminology.hl7.org/CodeSystem/encounter-type",
                    "http://terminology.hl7.org/CodeSystem/v2-0004",
                    "urn:oid:2.16.840.1.113883.4.642.3.248",
                    "http://snomed.info/sct",
                    # Cerner specific systems
                    "https://fhir.cerner.com/%/codeSet/71",
                    # EPIC specific systems
                    "urn:oid:1.2.840.114350.1.13.71.2.7.10.698084.10110",
                    "urn:oid:1.2.840.114350.1.13.71.2.7.10.698084.18875",
                    "urn:oid:1.2.840.114350.1.13.71.2.7.10.698084.30",
                    "urn:oid:1.2.840.114350.1.13.71.2.7.2.808267",
                ],
            ),
            EncConfig(
                column_hierarchy=[("servicetype", dict)],
                filter_priority=True,
                code_systems=[
                    "http://terminology.hl7.org/CodeSystem/service-type",
                    "http://snomed.info/sct",
                    # Cerner specific systems
                    "https://fhir.cerner.com/%/codeSet/34",
                    # EPIC specific systems
                    "urn:oid:2.16.840.1.113883.4.642.3.518",
                    "urn:oid:1.2.840.114350.1.13.71.2.7.10.698084.18886",
                ],
            ),
            EncConfig(
                column_hierarchy=[("priority", dict)],
                filter_priority=True,
                code_systems=[
                    "http://terminology.hl7.org/CodeSystem/v3-ActPriority",
                    "http://snomed.info/sct",
                    # Cerner specific systems
                    "https://fhir.cerner.com/%/codeSet/3",
                    # EPIC specific systems
                    "urn:oid:1.2.840.114350.1.13.71.2.7.10.698084.410",
                ],
            ),
            EncConfig(
                column_hierarchy=[("reasoncode", list)],
                filter_priority=True,
                code_systems=[
                    "http://terminology.hl7.org/CodeSystem/v3-ActPriority",
                    "http://snomed.info/sct",
                    "http://hl7.org/fhir/sid/icd-10-cm",
                    "http://hl7.org/fhir/sid/icd-9-cm",
                    # Cerner specific systems
                    "https://fhir.cerner.com/%/nomenclature",
                    # EPIC specific systems
                    "urn:oid:1.2.840.114350.1.13.71.2.7.2.728286",
                ],
            ),
            EncConfig(
                column_hierarchy=[
                    ("hospitalization", dict),
                    ("dischargedisposition", dict),
                ],
                expected={"dischargedisposition": sql_utils.CODEABLE_CONCEPT},
            ),
        ]
        self.queries += sql_utils.denormalize_complex_objects(database, code_configs)

    def prepare_queries(
        self,
        *args,
        config: cumulus_library.StudyConfig,
        **kwargs,
    ):
        self.denormalize_codes(config.db)
        validated_schema = sql_utils.validate_schema(config.db, expected_table_cols)
        self.queries += [
            core_templates.get_core_template("encounter", validated_schema),
            core_templates.get_core_template("incomplete_encounter", validated_schema),
        ]



import cumulus_library
from cumulus_library.studies.core.core_templates import core_templates
from cumulus_library.template_sql import sql_utils

expected_table_cols = {
    "procedure": {
        "id": [],
        "status": [],
        "subject": sql_utils.REFERENCE,
        "encounter": sql_utils.REFERENCE,
        "performedDateTime": [],
        "performedPeriod": ["start", "end"],
    }
}


class CoreProcedureBuilder(cumulus_library.BaseTableBuilder):
    display_text = "Creating Procedure tables..."

    def prepare_queries(self, *args, config: cumulus_library.StudyConfig, **kwargs):
        code_sources = [
            sql_utils.CodeableConceptConfig(
                source_table="procedure",
                column_hierarchy=[("category", dict)],
                target_table="core__procedure_dn_category",
            ),
            sql_utils.CodeableConceptConfig(
                source_table="procedure",
                column_hierarchy=[("code", dict)],
                target_table="core__procedure_dn_code",
            ),
        ]
        self.queries += sql_utils.denormalize_complex_objects(config.db, code_sources)
        validated_schema = sql_utils.validate_schema(config.db, expected_table_cols)
        self.queries.append(core_templates.get_core_template("procedure", validated_schema))



"""Class for loading study configuration data from manifest.toml files"""

import dataclasses
import pathlib
import re
import subprocess
import sys
import tomllib

from cumulus_library import errors


@dataclasses.dataclass(kw_only=True)
class ManifestExport:
    name: str
    export_type: str


class StudyManifest:
    """Handles interactions with study directories and manifest files"""

    PREFIX_REGEX = re.compile(r"[a-zA-Z][a-zA-Z0-9_]*")

    def __init__(
        self,
        study_path: pathlib.Path | None = None,
        data_path: pathlib.Path | None = None,
        *,
        options: dict[str, str] | None = None,
    ):
        """Instantiates a StudyManifest.

        :param study_path: A Path object pointing to the dir of the manifest, optional
        :param data_path: A Path object pointing to the dir to save/load ancillary files from,
            optional
        :param options: Command line study-specific options for dynamic manifest values, optional
        """
        self._study_prefix = None
        self._study_path = None
        self._study_config = {}
        if study_path is not None:
            self._load_study_manifest(study_path, options or {})
        self.data_path = data_path

    def __repr__(self):
        return str(self._study_config)

    ### toml parsing helper functions
    def _load_study_manifest(self, study_path: pathlib.Path, options: dict[str, str]) -> None:
        """Reads in a config object from a directory containing a manifest.toml

        :param study_path: A pathlib.Path object pointing to a study directory
        :param options: Command line study-specific options (--option=A:B)
        :raises StudyManifestParsingError: the manifest.toml is malformed or missing.
        """
        try:
            with open(f"{study_path}/manifest.toml", "rb") as file:
                config = tomllib.load(file)
        except FileNotFoundError as e:
            raise errors.StudyManifestFilesystemError(
                f"Missing or invalid manifest found at {study_path}"
            ) from e
        except tomllib.TOMLDecodeError as e:
            # just unify the error classes for convenience of catching them
            raise errors.StudyManifestParsingError(str(e)) from e

        self._study_config = config
        self._study_path = study_path

        if dynamic_study_prefix := config.get("dynamic_study_prefix"):
            self._study_prefix = self._run_dynamic_script(dynamic_study_prefix, options)
        elif config.get("study_prefix") and isinstance(config["study_prefix"], str):
            self._study_prefix = config["study_prefix"]

        if not self._study_prefix or not re.fullmatch(self.PREFIX_REGEX, self._study_prefix):
            raise errors.StudyManifestParsingError(f"Invalid prefix in manifest at {study_path}")
        self._study_prefix = self._study_prefix.lower()

    def get_study_prefix(self) -> str | None:
        """Reads the name of a study prefix from the in-memory study config

        :returns: A string of the prefix in the manifest, or None if not found
        """
        return self._study_prefix

    def get_dedicated_schema(self) -> str | None:
        """Reads the contents of the dedicated schema in the options dict

        :returns: A dictionary of objects, or None if not found
        """
        options = self._study_config.get("advanced_options", {})
        return options.get("dedicated_schema")

    def get_file_list(self, continue_from: str | None = None) -> list[str] | None:
        """Reads the contents of the file_config array from the manifest

        :returns: An array of files from the manifest, or None if not found.
        """
        config = self._study_config.get("file_config", {})
        files = config.get("file_names", []) or []
        if not files:
            files = (
                self.get_table_builder_file_list()
                + self.get_sql_file_list()
                + self.get_counts_builder_file_list()
                + self.get_statistics_file_list()
            )
        if continue_from:
            for pos, file in enumerate(files):
                if continue_from.split(".", 1)[0] == file.split(".", 1)[0]:
                    files = files[pos:]
                    break
            else:
                raise errors.StudyManifestParsingError(f"No files matching '{continue_from}' found")
        return files

    # The following four functions are considered deprecated, and can be removed
    # after we update studies to use the new methodology
    def get_sql_file_list(self, continue_from: str | None = None) -> list[str] | None:
        """Reads the contents of the sql_config array from the manifest

        :returns: An array of sql files from the manifest, or None if not found.
        """
        sql_config = self._study_config.get("sql_config", {})
        sql_files = sql_config.get("file_names", []) or []
        if continue_from:  # pragma: no cover
            for pos, file in enumerate(sql_files):
                if continue_from.replace(".sql", "") == file.replace(".sql", ""):
                    sql_files = sql_files[pos:]
                    break
            else:
                raise errors.StudyManifestParsingError(
                    f"No tables matching '{continue_from}' found"
                )
        return sql_files

    def get_table_builder_file_list(self) -> list[str] | None:
        """Reads the contents of the table_builder_config array from the manifest

        :returns: An array of sql files from the manifest, or None if not found.
        """
        sql_config = self._study_config.get("table_builder_config", {})
        return sql_config.get("file_names", [])

    def get_counts_builder_file_list(self) -> list[str] | None:
        """Reads the contents of the counts_builder_config array from the manifest

        :returns: An array of sql files from the manifest, or None if not found.
        """
        sql_config = self._study_config.get("counts_builder_config", {})
        return sql_config.get("file_names", [])

    def get_statistics_file_list(self) -> list[str] | None:
        """Reads the contents of the statistics_config array from the manifest

        :returns: An array of statistics toml files from the manifest,
          or None if not found.
        """
        stats_config = self._study_config.get("statistics_config", {})
        return stats_config.get("file_names", [])

    # End of deprecated section

    def get_export_table_list(self) -> list[ManifestExport] | None:
        """Reads the contents of the export_list array from the manifest

        :returns: An array of tuples (table, export type) to export from the manifest,
        or None if not found.
        """
        export_config = self._study_config.get("export_config", {})
        export_table_list = []

        for section in [
            ("export_list", "cube"),
            ("count_list", "cube"),
            ("flat_list", "flat"),
            ("meta_list", "meta"),
            ("annotated_count_list", "annotated_cube"),
        ]:
            section_list = export_config.get(section[0], []) or []
            for table in section_list:
                if table.startswith(f"{self.get_study_prefix()}__"):
                    export_table_list.append(ManifestExport(name=table, export_type=section[1]))
                elif "__" in table:  # has a prefix, just the wrong one
                    raise errors.StudyManifestParsingError(
                        f"{table} in export list does not start with prefix "
                        f"{self.get_study_prefix()}__ - check your manifest file."
                    )
                else:
                    # Add the prefix for them (helpful in dynamic prefix cases where the prefix
                    # is not known ahead of time)
                    export_table_list.append(
                        ManifestExport(
                            name=f"{self.get_study_prefix()}__{table}", export_type=section[1]
                        )
                    )
        found_name = set()
        for export in export_table_list:
            if export.name in found_name:
                raise errors.StudyManifestParsingError(
                    f"Table {export.name} is defined in multiple export sections. Tables should "
                    "only be defined in one export section in the study manifest"
                )
            else:
                found_name.add(export.name)
        return export_table_list

    def get_all_files(self, file_type: str):
        """Convenience method for getting files of a type from a manifest"""
        files = self.get_file_list()
        return [file for file in files if file.endswith(file_type)]

    def get_all_generators(self) -> list[str]:
        """Convenience method for getting builder-based files"""
        return self.get_all_files(".py")

    def get_all_workflows(self) -> list[str]:
        """Convenience method for getting workflow config files"""
        return self.get_all_files(".toml")

    def get_prefix_with_seperator(self) -> str:
        """Convenience method for getting the appropriate prefix for tables"""
        if dedicated := self.get_dedicated_schema():
            return f"{dedicated}."
        return f"{self.get_study_prefix()}__"

    ### Dynamic Python code support

    def _run_dynamic_script(self, filename: str, options: dict[str, str]) -> str:
        if not sys.executable:
            raise RuntimeError("Unknown Python executable, dynamic manifest values not supported")

        full_path = f"{self._study_path}/{filename}"
        option_args = [f"--{key}={value}" for key, value in options.items()]
        result = subprocess.run(  # noqa: S603
            [sys.executable, full_path, *option_args],
            check=True,
            capture_output=True,
        )

        return result.stdout.decode("utf8").strip()



"""Builder for generating subsets of RxNorm data from a given valueset"""

import pathlib

from cumulus_library import BaseTableBuilder, base_utils, study_manifest
from cumulus_library.builders.valueset import umls, valueset_utils
from cumulus_library.template_sql import base_templates


class RxNormValuesetBuilder(BaseTableBuilder):
    display_text = "Creating RxNorm subsets from valueset definitions..."
    base_path = pathlib.Path(__file__).resolve().parent

    def prepare_queries(
        self,
        *args,
        config: base_utils.StudyConfig,
        manifest: study_manifest.StudyManifest,
        valueset_config: valueset_utils.ValuesetConfig,
        **kwargs,
    ):
        study_prefix = f"{manifest.get_study_prefix()}__"
        table_prefix = ""
        if valueset_config.table_prefix:
            table_prefix = f"{valueset_config.table_prefix}_"

        def get_create_view_filter_by(
            a_table: str,
            *,
            columns: list,
            b_table: str | None = None,
            a_join_col: str | None = None,
            b_join_col: str | None = None,
            a_schema: str | None = None,
            view_name: str | None = None,
            join_clauses: list | None = None,
            column_aliases: dict | None = None,
        ):
            a_schema = a_schema or "rxnorm."
            a_join_col = a_join_col or "a.rxcui"
            b_join_col = b_join_col or "b.rxcui"
            b_table = b_table or f"{study_prefix}{table_prefix}valuesets"
            join_clauses = join_clauses or [f"{a_join_col} = {b_join_col}"]
            view_name = view_name or f"{study_prefix}{table_prefix}{a_table}"

            return base_templates.get_create_view_from_tables(
                view_name=view_name,
                tables=[f"{a_schema}{a_table}", b_table],
                table_aliases=["a", "b"],
                join_clauses=join_clauses,
                columns=columns,
                column_aliases=column_aliases,
                distinct=True,
            )

        if valueset_config.keyword_file is not None:
            with open(manifest._study_path / valueset_config.keyword_file) as f:
                keywords = [row.rstrip() for row in f.readlines()]

            self.queries.append(
                base_templates.get_base_template(
                    "create_keyword_annotated_table",
                    self.base_path / "template_sql",
                    keywords=keywords,
                    source_table="rxnorm.rxnconso",
                    table_name=f"{study_prefix}{table_prefix}all_rxnconso_keywords",
                )
            )
        umls.generate_umls_tables(config, manifest, valueset_config)
        self.queries.append(
            get_create_view_filter_by(
                "rxnconso",
                b_table=f"{study_prefix}{table_prefix}vsac_valuesets",
                columns=["a.rxcui", "a.str", "a.tty", "a.sab", "a.code", "b.steward"],
                view_name=f"{study_prefix}{table_prefix}vsac_valuesets_hydrated",
            )
        )
        self.queries.append(
            base_templates.get_create_table_from_union(
                table_name=f"{study_prefix}{table_prefix}valuesets",
                tables=[
                    f"{study_prefix}{table_prefix}umls_valuesets",
                    f"{study_prefix}{table_prefix}vsac_valuesets_hydrated",
                ],
                columns=["rxcui", "str", "tty", "sab", "code", "steward"],
            )
        )
        self.queries.append(
            get_create_view_filter_by(
                "rxnconso",
                columns=["a.rxcui", "a.str", "a.tty", "a.sab", "a.code", "b.steward"],
            )
        )
        if valueset_config.keyword_file is not None:
            self.queries.append(
                base_templates.get_base_template(
                    "create_keyword_annotated_table",
                    self.base_path / "template_sql",
                    keywords=keywords,
                    source_table=f"{study_prefix}{table_prefix}rxnconso",
                    table_name=f"{study_prefix}{table_prefix}rxnconso_keywords",
                )
            )
        self.queries.append(
            get_create_view_filter_by(
                "rxnsty",
                columns=["a.rxcui", "a.tui", "a.stn", "a.sty", "a.atui", "a.cvf"],
            )
        )
        self.queries.append(
            get_create_view_filter_by(
                "rxnrel",
                columns=[
                    "a.rxcui1",
                    "a.rxaui1",
                    "a.stype1",
                    "a.rel",
                    "a.rxcui2",
                    "a.rxaui2",
                    "a.stype2",
                    "a.rela",
                    "a.rui",
                    "a.srui",
                    "a.sab",
                    "a.sl",
                    "a.rg",
                    "b.steward",
                ],
                a_join_col="a.rxcui1",
            )
        )
        self.queries.append(
            get_create_view_filter_by(
                f"{table_prefix}rxnconso_keywords",
                view_name=f"{study_prefix}{table_prefix}rela",
                a_schema=f"{study_prefix}",
                b_table=f"{study_prefix}{table_prefix}rxnrel",
                columns=[
                    "a.rxcui",
                    "a.str",
                    "a.tty",
                    "a.sab",
                    "b.rxcui2",
                    "b.rel",
                    "b.rela",
                    "b.rui",
                    "b.steward",
                ],
                b_join_col="b.rxcui1",
            )
        )



"""Database abstraction layer for AWS Athena

See base.py for design rules of thumb - notably, don't use non-PEP249 database
features, like pyathena's async cursors, to simplify cross-db behavior.
"""

import collections
import os
import pathlib

import awswrangler
import boto3
import botocore
import numpy
import pandas
import pyathena
from pyathena.common import BaseCursor as AthenaCursor
from pyathena.pandas.cursor import PandasCursor as AthenaPandasCursor

from cumulus_library import errors
from cumulus_library.databases import base


class AthenaDatabaseBackend(base.DatabaseBackend):
    """Database backend that can talk to AWS Athena"""

    def __init__(self, region: str, work_group: str, profile: str, schema_name: str):
        super().__init__(schema_name)

        self.db_type = "athena"
        self.region = region
        self.work_group = work_group
        self.profile = profile
        self.schema_name = schema_name
        self.connection = None
        self.connect_kwargs = {}

    def init_errors(self):  # pragma: no cover
        return ["COLUMN_NOT_FOUND", "TABLE_NOT_FOUND"]

    def connect(self):
        # the profile may not be required, provided the above three AWS env vars
        # are set. If both are present, the env vars take precedence
        if self.profile is not None:
            self.connect_kwargs["profile_name"] = self.profile
        for aws_env_name in [
            "AWS_ACCESS_KEY_ID",
            "AWS_SECRET_ACCESS_KEY",
            "AWS_SESSION_TOKEN",
        ]:
            if aws_env_val := os.environ.get(aws_env_name):
                self.connect_kwargs[aws_env_name.lower()] = aws_env_val

        self.connection = pyathena.connect(
            region_name=self.region,
            work_group=self.work_group,
            schema_name=self.schema_name,
            **self.connect_kwargs,
        )

    def cursor(self) -> AthenaCursor:
        return self.connection.cursor()

    def pandas_cursor(self) -> AthenaPandasCursor:
        return self.connection.cursor(cursor=AthenaPandasCursor)

    def execute_as_pandas(
        self, sql: str, chunksize: int | None = None
    ) -> (pandas.DataFrame | collections.abc.Iterator[pandas.DataFrame], list[tuple]):
        query = self.pandas_cursor().execute(sql, chunksize=chunksize)
        return query.as_pandas(), query.description

    def parser(self) -> base.DatabaseParser:
        return AthenaParser()

    def operational_errors(self) -> tuple[type[Exception], ...]:
        return (pyathena.OperationalError,)

    def col_parquet_types_from_pandas(self, field_types: list) -> list:
        output = []
        for field in field_types:
            match field:
                case numpy.dtypes.ObjectDType():
                    output.append("STRING")
                case pandas.core.arrays.integer.Int64Dtype() | numpy.dtypes.Int64DType():
                    output.append("INT")
                case numpy.dtypes.Float64DType():
                    output.append("DOUBLE")
                case numpy.dtypes.BoolDType():
                    output.append("BOOLEAN")
                case numpy.dtypes.DateTime64DType():
                    output.append("TIMESTAMP")
                case _:
                    raise errors.CumulusLibraryError(
                        f"Unsupported pandas type {type(field)} found."
                    )
        return output

    def upload_file(
        self,
        *,
        file: pathlib.Path,
        study: str,
        topic: str,
        remote_filename: str | None = None,
        force_upload=False,
    ) -> str | None:
        # We'll investigate the connection to get the relevant S3 upload path.
        workgroup = self.connection._client.get_work_group(WorkGroup=self.work_group)
        wg_conf = workgroup["WorkGroup"]["Configuration"]["ResultConfiguration"]
        s3_path = wg_conf["OutputLocation"]
        bucket = "/".join(s3_path.split("/")[2:3])
        key_prefix = "/".join(s3_path.split("/")[3:])
        encryption_type = wg_conf.get("EncryptionConfiguration", {}).get("EncryptionOption")
        if encryption_type is None:
            raise errors.AWSError(
                f"The workgroup {self.work_group} is not reporting an encrypted status. "
                f"Either bucket {bucket} is not encrypted, or the workgroup is overriding the "
                "bucket's default encryption. Please confirm the encryption status of both. "
                "Cumulus requires encryption as part of ensuring safety for limited data sets"
            )
        kms_arn = wg_conf.get("EncryptionConfiguration", {}).get("KmsKey", None)
        s3_key = f"{key_prefix}cumulus_user_uploads/{self.schema_name}/{study}/{topic}"
        if not remote_filename:
            remote_filename = file.name

        session = boto3.Session(profile_name=self.connection.profile_name)
        s3_client = session.client("s3")
        if not force_upload:
            res = s3_client.list_objects_v2(
                Bucket=bucket,
                Prefix=f"{s3_key}/{remote_filename}",
            )
            if res["KeyCount"] > 0:
                return f"s3://{bucket}/{s3_key}"
        with open(file, "rb") as b_file:
            s3_client.put_object(
                Bucket=bucket,
                Key=f"{s3_key}/{remote_filename}",
                Body=b_file,
                ServerSideEncryption="aws:kms",
                SSEKMSKeyId=kms_arn,
            )
        return f"s3://{bucket}/{s3_key}"

    def _clean_bucket_path(self, client, bucket, res):
        for file in res["Contents"]:
            client.delete_object(Bucket=bucket, Key=file["Key"])

    def export_table_as_parquet(
        self, table_name: str, file_name: str, location: pathlib.Path, *args, **kwargs
    ) -> bool:
        session = boto3.session.Session(
            **self.connect_kwargs,
        )
        s3_client = session.client("s3")
        workgroup = self.connection._client.get_work_group(WorkGroup=self.work_group)
        wg_conf = workgroup["WorkGroup"]["Configuration"]["ResultConfiguration"]
        s3_path = wg_conf["OutputLocation"]
        bucket = "/".join(s3_path.split("/")[2:3])
        output_path = location / f"{file_name}"
        s3_path = f"s3://{bucket}/export/{file_name}"
        # Cleanup location in case there was an error of some kind
        res = s3_client.list_objects_v2(Bucket=bucket, Prefix=f"export/{file_name}")
        if "Contents" in res:
            self._clean_bucket_path(s3_client, bucket, res)

        self.connection.cursor().execute(f"""UNLOAD
            (SELECT * FROM {table_name})
            TO '{s3_path}'
            WITH (format='PARQUET', compression='SNAPPY')
            """)  # noqa: S608
        # UNLOAD is not guaranteed to create a single file. AWS Wrangler's read_parquet
        # allows us to ignore that wrinkle
        try:
            df = awswrangler.s3.read_parquet(s3_path, boto3_session=session)
        except awswrangler.exceptions.NoFilesFound:
            return False
        df = df.sort_values(by=list(df.columns), ascending=False, na_position="first")
        df.to_parquet(output_path)
        res = s3_client.list_objects_v2(Bucket=bucket, Prefix=f"export/{file_name}")
        self._clean_bucket_path(s3_client, bucket, res)
        return True

    def create_schema(self, schema_name) -> None:
        """Creates a new schema object inside the database"""
        glue_client = boto3.client("glue")
        try:
            glue_client.get_database(Name=schema_name)
        except botocore.exceptions.ClientError:
            glue_client.create_database(DatabaseInput={"Name": schema_name})

    def close(self) -> None:
        if self.connection is not None:  # pragma: no cover
            self.connection.close()


class AthenaParser(base.DatabaseParser):
    def _find_type_len(self, row: str) -> int:
        """Finds the end of a type string like row(...) or array(row(...))"""
        # Note that this assumes the string is well formatted.
        depth = 0
        for i in range(len(row)):
            match row[i]:
                case ",":
                    if depth == 0:
                        break
                case "(":
                    depth += 1
                case ")":
                    depth -= 1
        return i

    def _split_row(self, row: str) -> dict[str, str]:
        # Must handle "name type, name row(...), name type, name row(...)"
        result = {}
        # Real athena doesn't add extra spaces, but our unit tests do for
        # readability, so let's strip out whitespace as we parse.
        while row := row.strip():
            name, remainder = row.split(" ", 1)
            type_len = self._find_type_len(remainder)
            result[name] = remainder[0:type_len]
            row = remainder[type_len + 2 :]  # skip comma and space
        return result

    def parse_found_schema(self, schema: dict[str, str]) -> dict:
        # A sample response for table `observation`, column `component`:
        #   array(row(code varchar, display varchar)),
        #             text varchar, id varchar)
        parsed = {}

        for key, value in schema.items():
            # Strip arrays out, they don't affect the shape of our schema.
            while value.startswith("array("):
                value = value.removeprefix("array(")
                value = value.removesuffix(")")

            if value.startswith("row("):
                value = value.removeprefix("row(")
                value = value.removesuffix(")")
                parsed[key] = self.parse_found_schema(self._split_row(value))
            else:
                parsed[key] = {}

        return parsed



"""Module for extracting US core extensions from patient records"""

import cumulus_library
from cumulus_library import databases
from cumulus_library.studies.core.core_templates import core_templates
from cumulus_library.template_sql import base_templates, sql_utils

expected_table_cols = {
    "patient": {
        "id": [],
        "gender": [],
        "address": {"postalCode": {}},
        "birthDate": [],
    }
}


class PatientBuilder(cumulus_library.BaseTableBuilder):
    display_text = "Creating Patient tables..."

    @staticmethod
    def make_extension_query(
        database: databases.DatabaseBackend,
        name: str,
        url: str,
    ) -> str:
        has_extensions = sql_utils.is_field_present(
            database=database,
            source_table="patient",
            source_col="extension",
            expected={
                "extension": {
                    "url": {},
                    "valueCoding": sql_utils.CODING,
                },
                "url": {},
            },
        )
        if has_extensions:
            config = sql_utils.ExtensionConfig(
                source_table="patient",
                source_id="id",
                target_table=f"core__patient_ext_{name}",
                target_col_prefix=name,
                fhir_extension=url,
                ext_systems=["ombCategory", "detailed"],
                is_array=True,
            )
            return base_templates.get_extension_denormalize_query(config)
        else:
            return base_templates.get_ctas_empty_query(
                schema_name=database.schema_name,
                table_name=f"core__patient_ext_{name}",
                table_cols=["id", "system", f"{name}_code", f"{name}_display"],
            )

    def prepare_queries(
        self,
        *args,
        config: cumulus_library.StudyConfig,
        **kwargs,
    ):
        """constructs queries related to patient extensions of interest

        :param config: A study config object
        """
        extension_types = [
            {
                "name": "race",
                "fhirpath": "http://hl7.org/fhir/us/core/StructureDefinition/us-core-race",
            },
            {
                "name": "ethnicity",
                "fhirpath": "http://hl7.org/fhir/us/core/StructureDefinition/us-core-ethnicity",
            },
        ]
        for extension in extension_types:
            self.queries.append(
                self.make_extension_query(config.db, extension["name"], extension["fhirpath"])
            )

        validated_schema = sql_utils.validate_schema(config.db, expected_table_cols)
        self.queries.append(core_templates.get_core_template("patient", validated_schema))



import argparse
import json
import pathlib

import pandas
import rich

from cumulus_library import base_utils
from cumulus_library.apis import umls


def download_oid_data(
    *,
    steward: str,
    oid: str,
    api_key: str | None = None,
    config: base_utils.StudyConfig | None = None,
    force_upload: str | None = None,
    path: pathlib.Path | None = None,
) -> bool:
    """Fetches code definitions (assumed to be RXNorm coded) from VSAC
    :keyword steward: the human readable label for the valueset
    :keyword oid: the ID of the valueset in the VSAC database
    :keyword config: A StudyConfig object. If umls_key is none, will check the
        UMLS_API_KEY env var
    :keyword path: A path to write data to (default: ../data)
    :returns: True if file created, false otherwise (mostly for testing)
    """
    if config:
        api_key = config.umls_key
        force_upload = config.force_upload
    if not path:
        path = pathlib.Path(__file__).parent.parent / "data"  # pragma: no cover
    path.mkdir(exist_ok=True, parents=True)
    if not (force_upload) and (path / f"{steward}.parquet").exists():
        rich.print(f"{steward} data present at {path}, skipping download.")
        return False
    rich.print(f"Downloading {steward} to {path}")
    api = umls.UmlsApi(api_key=api_key or api_key)
    output = []

    response = api.get_vsac_valuesets(oid=oid)
    for valueset in response:
        contains = valueset.get("expansion").get("contains", [])
        for data in contains:
            output.append([data["code"], data["display"]])
    output_df = pandas.DataFrame(output, columns=["RXCUI", "STR"])
    output_df.to_csv(path / f"{steward}.tsv", index=False, header=False, sep="\t")
    output_df.to_parquet(path / f"{steward}.parquet")
    with open(path / f"{steward}.json", "w") as f:
        f.write(json.dumps(response))
    return True


def main(cli_args=None):
    """Deprecated CLI interface"""
    parser = argparse.ArgumentParser()
    (
        parser.add_argument(
            "--steward", help="Human-friendly name for steward (used for filenames)", default=None
        ),
    )
    parser.add_argument("--oid", help="oid to look up codes for", default=None)
    parser.add_argument("--api-key", help="UMLS api key", default=None)
    parser.add_argument(
        "--force-upload",
        help="Force redownloading of data even if it already exists",
        action="store_true",
    )
    parser.add_argument("--path", help="optional path to write data to", default=None)
    args = parser.parse_args(cli_args)
    return download_oid_data(
        steward=args.steward,
        oid=args.oid,
        api_key=args.api_key,
        force_upload=args.force_upload,
        path=pathlib.Path(args.path),
    )


if __name__ == "__main__":
    main()  # pragma: no cover



# Module for generating Propensity Score matching cohorts

import json
import os
import pathlib
import sys
import tomllib
import warnings
from dataclasses import dataclass

import matplotlib.pyplot as plt
import pandas
import seaborn as sns
from psmpy import PsmPy

# these imports are mimicing PsmPy imports for re-implemented functions
from psmpy.functions import cohenD

from cumulus_library import BaseTableBuilder, base_utils, databases
from cumulus_library.builders.statistics_templates import psm_templates
from cumulus_library.template_sql import base_templates


@dataclass
class PsmConfig:
    """Provides expected values for PSM execution

    These values should be read in from a toml configuration file.
    See docs/statistics/propensity-score-matching.md for an example with details about
    the expected values for these fields.

    A word of caution about sampling: the assumptions around PSM analysis require
    that any sampling should not use replacement, so do not turn on panda's dataframe
    replacement. This will mean that very small population sizes (i.e. < 20ish)
    may cause errors to be generated.
    """

    classification_json: str
    pos_source_table: str
    neg_source_table: str
    target_table: str
    primary_ref: str
    count_ref: str
    count_table: str
    dependent_variable: str
    pos_sample_size: int
    neg_sample_size: int
    join_cols_by_table: dict[str, dict]
    seed: int


class PsmBuilder(BaseTableBuilder):
    """TableBuilder for creating PSM tables"""

    display_text = "Building PSM tables..."

    def __init__(self, toml_config_path: str, data_path: pathlib.Path, **kwargs):
        """Loads PSM job details from a PSM configuration file"""
        super().__init__()
        # We're stashing the toml path for error reporting later
        self.toml_path = toml_config_path
        self.data_path = data_path
        try:
            with open(self.toml_path, "rb") as file:
                toml_config = tomllib.load(file)

        except OSError:
            sys.exit(f"PSM configuration not found at {self.toml_path}")
        try:
            toml_dir = pathlib.Path(self.toml_path).parent
            self.config = PsmConfig(
                classification_json=str(toml_dir.joinpath(toml_config["classification_json"])),
                pos_source_table=toml_config["pos_source_table"],
                neg_source_table=toml_config["neg_source_table"],
                target_table=toml_config["target_table"],
                primary_ref=toml_config["primary_ref"],
                dependent_variable=toml_config["dependent_variable"],
                pos_sample_size=toml_config["pos_sample_size"],
                neg_sample_size=toml_config["neg_sample_size"],
                join_cols_by_table=toml_config.get("join_cols_by_table", {}),
                count_ref=toml_config.get("count_ref", None),
                count_table=toml_config.get("count_table", None),
                seed=toml_config.get("seed", 123),
            )
        except KeyError:
            sys.exit(
                f"PSM configuration {toml_config_path} contains missing/invalid keys."
                "Check the PSM documentation for an example config with more details:\n"
                "https://docs.smarthealthit.org/cumulus/library/statistics/propensity-score-matching.html"
            )

    def _get_symptoms_dict(self, path: str) -> dict:
        """convenience function for loading symptoms dictionaries from a json file"""
        with open(path, encoding="UTF-8") as f:
            symptoms = json.load(f)
        return symptoms

    def _get_sampled_ids(
        self,
        cursor: databases.DatabaseCursor,
        schema: str,
        query: str,
        sample_size: int,
        dependent_variable: str,
        is_positive: bool,
    ):
        """Creates a table containing randomly sampled patients for PSM analysis

        To use this, it is assumed you have already identified a cohort of positively
        IDed patients as a manual process.
        :param cursor: A valid DatabaseCusror:
        :param schema: the schema/database name where the data exists
        :param query: a query generated from the psm_dsitinct_ids template
        :param sample_size: the number of records to include in the random sample.
            This should generally be >= 20.
        :param dependent_variable: the name to use for your filtering column
        :param is_positive: defines the value to be used for your filtering column
        """
        df = cursor.execute(query).as_pandas()
        df = df.sort_values(by=[self.config.primary_ref])
        df = (
            df.sample(n=sample_size, random_state=self.config.seed)
            .sort_values(by=[self.config.primary_ref])
            .reset_index()
            .drop("index", axis=1)
        )

        df[dependent_variable] = is_positive
        return df

    def _create_covariate_table(
        self, cursor: databases.DatabaseCursor, schema: str, table_suffix: str
    ):
        """Creates a covariate table from the loaded toml config"""
        # checks for primary & link ref being the same
        source_refs = list({self.config.primary_ref, self.config.count_ref} - {None})
        pos_query = psm_templates.get_distinct_ids(source_refs, self.config.pos_source_table)
        pos = self._get_sampled_ids(
            cursor,
            schema,
            pos_query,
            self.config.pos_sample_size,
            self.config.dependent_variable,
            1,
        )
        neg_query = psm_templates.get_distinct_ids(
            source_refs,
            self.config.neg_source_table,
            join_id=self.config.primary_ref,
            filter_table=self.config.pos_source_table,
        )
        neg = self._get_sampled_ids(
            cursor,
            schema,
            neg_query,
            self.config.neg_sample_size,
            self.config.dependent_variable,
            0,
        )

        cohort = pandas.concat([pos, neg])
        ctas_query = base_templates.get_ctas_query_from_df(
            schema,
            f"{self.config.pos_source_table}_sampled_ids_{table_suffix}",
            cohort,
        )
        self.queries.append(ctas_query)

        dataset_query = psm_templates.get_create_covariate_table(
            target_table=f"{self.config.target_table}_{table_suffix}",
            pos_source_table=self.config.pos_source_table,
            neg_source_table=self.config.neg_source_table,
            table_suffix=table_suffix,
            primary_ref=self.config.primary_ref,
            dependent_variable=self.config.dependent_variable,
            join_cols_by_table=self.config.join_cols_by_table,
            count_ref=self.config.count_ref,
            count_table=self.config.count_table,
        )
        self.queries.append(dataset_query)

    def psm_plot_match(
        self,
        psm,
        matched_entity="propensity_logit",
        Title="Side by side matched controls",
        Ylabel="Number of patients",
        Xlabel="propensity logit",
        names=None,
        colors=None,
        save=True,
        filename="propensity_match.png",
    ):
        """Plots knn match data

        This function re-implements psm.plot_match, with the only changes
        allowing for specifiying a filename/location for saving plots to,
        and passing in the psm object instead of assuming a call from inside
        the PsmPy class.
        """
        names = names or ["positive_cohort", "negative_cohort"]
        colors = colors or ["#E69F00", "#56B4E9"]
        dftreat = psm.df_matched[psm.df_matched[psm.treatment] == 1]
        dfcontrol = psm.df_matched[psm.df_matched[psm.treatment] == 0]
        x1 = dftreat[matched_entity]
        x2 = dfcontrol[matched_entity]
        colors = colors
        names = names
        sns.set_style("white")
        plt.hist([x1, x2], color=colors, label=names)
        plt.legend()
        plt.xlabel(Xlabel)
        plt.ylabel(Ylabel)
        plt.title(Title)
        if save:
            plt.savefig(filename, dpi=250)

    def psm_effect_size_plot(
        self,
        psm,
        title="Standardized Mean differences accross covariates before and after matching",
        before_color="#FCB754",
        after_color="#3EC8FB",
        save=False,
        filename="effect_size.png",
    ):
        """Plots effect size of variables for positive/negative matches

        This function re-implements psm.effect_size_plot, with the only changes
        allowing for specifiying a filename/location for saving plots to,
        and passing in the psm object instead of assuming a call from inside
        the PsmPy class.
        """
        df_preds_after = psm.df_matched[[psm.treatment] + psm.xvars]  # noqa: RUF005
        df_preds_b4 = psm.data[[psm.treatment] + psm.xvars]  # noqa: RUF005
        df_preds_after_float = df_preds_after.astype(float)
        df_preds_b4_float = df_preds_b4.astype(float)

        data = []
        for cl in psm.xvars:
            data.append([cl, "before", cohenD(df_preds_b4_float, psm.treatment, cl)])
            data.append([cl, "after", cohenD(df_preds_after_float, psm.treatment, cl)])
        psm.effect_size = pandas.DataFrame(data, columns=["Variable", "matching", "Effect Size"])
        sns.set_style("white")
        sns_plot = sns.barplot(
            data=psm.effect_size,
            y="Variable",
            x="Effect Size",
            hue="matching",
            palette=[before_color, after_color],
            orient="h",
        )
        sns_plot.set(title=title)
        if save:
            sns_plot.figure.savefig(filename, dpi=250, bbox_inches="tight")

    def generate_psm_analysis(
        self, cursor: databases.DatabaseCursor, schema: str, table_suffix: str
    ):
        """Runs PSM statistics on generated tables"""
        stats_table = f"{self.config.target_table}_{table_suffix}"
        cursor.execute(base_templates.get_alias_table_query(stats_table, self.config.target_table))
        df = cursor.execute(
            base_templates.get_select_all_query(self.config.target_table)
        ).as_pandas()
        symptoms_dict = self._get_symptoms_dict(self.config.classification_json)
        for dependent_variable, codes in symptoms_dict.items():
            df[dependent_variable] = df["code"].apply(lambda x: 1 if x in codes else 0)
        df = df.drop(columns="code")
        # instance_count present but unused for PSM if table contains a count_ref input
        # (it's intended for manual review)
        df = df.drop(columns="instance_count", errors="ignore")

        columns = []
        if self.config.join_cols_by_table is not None:
            for table_config in self.config.join_cols_by_table.values():
                for column in table_config["included_cols"]:
                    # If there are two elements, it's a SQL column that has been
                    # aliased, so we'll look for the alias name
                    if len(column) == 2:
                        columns.append(column[1])
                    # If there is one element, it's a straight SQL column we can
                    # use with no modification
                    elif len(column) == 1:
                        columns.append(column[0])
                    else:
                        sys.exit(
                            f"PSM config at {self.toml_path} contains an "
                            f"unexpected SQL column definition: {column}."
                            "Check the PSM documentation for valid usages."
                        )

        # This code block is replacing a column which may contain several categories
        # (like male/female/other/unknown for AdministrativeGender), and converts
        # it into a series of 1-hot columns for each distinct value in that column,
        for column in columns:
            encoded_df = pandas.get_dummies(df[column])
            df = pandas.concat([df, encoded_df], axis=1)
            df = df.drop(column, axis=1)
        try:
            psm = PsmPy(
                df,
                treatment=self.config.dependent_variable,
                indx=self.config.primary_ref,
                exclude=[],
            )

            # we expect psmpy to autodrop non-matching values, so we'll surpress it
            # mentioning workarounds for this behavior.
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=UserWarning)
                # This function populates the psm.predicted_data element, which is
                # required for things like the knn_matched() function call
                psm.logistic_ps(balance=True)
                # This function populates the psm.df_matched element
                psm.knn_matched(
                    matcher="propensity_logit",
                    replacement=False,
                    caliper=None,
                    drop_unmatched=True,
                )
                os.makedirs(self.data_path, exist_ok=True)
                self.psm_plot_match(
                    psm,
                    save=True,
                    filename=self.data_path / f"{stats_table}_propensity_match.png",
                )
                self.psm_effect_size_plot(
                    psm,
                    save=True,
                    filename=self.data_path / f"{stats_table}_effect_size.png",
                )
        except ZeroDivisionError:
            sys.exit(
                "Encountered a divide by zero error during statistical graph "
                "generation. Try increasing your sample size."
            )
        except ValueError:
            sys.exit(
                "Encountered a value error during KNN matching. Try increasing your sample size."
            )

    def prepare_queries(self, config: base_utils.StudyConfig, *args, table_suffix: str, **kwargs):
        self._create_covariate_table(config.db.cursor(), config.schema, table_suffix)

    def post_execution(
        self,
        config: base_utils.StudyConfig,
        *args,
        table_suffix: str | None = None,
        **kwargs,
    ):
        # super().execute_queries(cursor, schema, verbose, drop_table)
        self.generate_psm_analysis(config.db.cursor(), config.schema, table_suffix)



import cumulus_library
from cumulus_library.studies.core.core_templates import core_templates
from cumulus_library.template_sql import sql_utils

expected_table_cols = {
    "documentreference": {
        "id": [],
        "status": [],
        "date": [],
        "docStatus": [],
        "subject": sql_utils.REFERENCE,
        "context": {"encounter": sql_utils.REFERENCE, "period": ["start"]},
    }
}


class CoreDocumentreferenceBuilder(cumulus_library.BaseTableBuilder):
    display_text = "Creating DocumentReference tables..."

    def prepare_queries(
        self,
        *args,
        config: cumulus_library.StudyConfig,
        **kwargs,
    ):
        self.queries = sql_utils.denormalize_complex_objects(
            config.db,
            [
                sql_utils.CodeableConceptConfig(
                    source_table="documentreference",
                    source_id="id",
                    column_hierarchy=[("type", dict)],
                    target_table="core__documentreference_dn_type",
                ),
                # TODO: The US core profile allows an extensible code for category, but
                # it's unclear what the possible extended codes would be. For the time
                # being, we select the spec's preferred code system (which is a set of
                # one value, 'clinical-note').
                # It may be worth confirming the values in this field with either the
                # quality or discovery studies on an ongoing basis to find other uses
                # for this field.
                sql_utils.CodeableConceptConfig(
                    source_table="documentreference",
                    source_id="id",
                    column_hierarchy=[("category", list)],
                    filter_priority=True,
                    target_table="core__documentreference_dn_category",
                    code_systems=[
                        "http://hl7.org/fhir/us/core/ValueSet/us-core-documentreference-category"
                    ],
                ),
                sql_utils.CodingConfig(
                    source_table="documentreference",
                    source_id="id",
                    column_hierarchy=[("content", list), ("format", dict)],
                    target_table="core__documentreference_dn_format",
                    expected={"format": sql_utils.CODING},
                ),
            ],
        )
        validated_schema = sql_utils.validate_schema(config.db, expected_table_cols)
        self.queries.append(core_templates.get_core_template("documentreference", validated_schema))



# A collection of codes & codeableConcepts to extract available codes from.

from cumulus_library.template_sql import sql_utils

code_list = [
    # AllergyIntolerance
    {
        "table_name": "allergyintolerance",
        "column_hierarchy": [("clinicalstatus", dict), ("coding", list)],
    },
    {
        "table_name": "allergyintolerance",
        "column_hierarchy": [("verificationstatus", dict), ("coding", list)],
    },
    {
        "table_name": "allergyintolerance",
        "column_hierarchy": [("code", dict), ("coding", list)],
    },
    {
        "table_name": "allergyintolerance",
        "column_hierarchy": [("reaction", list), ("substance", dict), ("coding", list)],
    },
    {
        "table_name": "allergyintolerance",
        "column_hierarchy": [("reaction", list), ("manifestation", list), ("coding", list)],
    },
    {
        "table_name": "allergyintolerance",
        "column_hierarchy": [("reaction", list), ("exposureroute", dict), ("coding", list)],
    },
    # Condition
    {
        "table_name": "condition",
        "column_hierarchy": [("category", list), ("coding", list)],
    },
    {
        "table_name": "condition",
        "column_hierarchy": [("code", dict), ("coding", list)],
    },
    # DiagnosticReport
    {
        "table_name": "diagnosticreport",
        "column_hierarchy": [("category", list), ("coding", list)],
    },
    {
        "table_name": "diagnosticreport",
        "column_hierarchy": [("code", dict), ("coding", list)],
    },
    {
        "table_name": "diagnosticreport",
        "column_hierarchy": [("conclusioncode", list), ("coding", list)],
    },
    # DocumentReference
    {
        "table_name": "documentreference",
        "column_hierarchy": [("type", dict), ("coding", list)],
    },
    {
        "table_name": "documentreference",
        "column_hierarchy": [("category", list), ("coding", list)],
    },
    # Encounter
    {
        "table_name": "encounter",
        "column_hierarchy": [("class", dict)],
        "expected": sql_utils.CODING,
    },
    {
        "table_name": "encounter",
        "column_hierarchy": [("type", list), ("coding", list)],
    },
    {
        "table_name": "encounter",
        "column_hierarchy": [("servicetype", dict), ("coding", list)],
    },
    {
        "table_name": "encounter",
        "column_hierarchy": [("priority", dict), ("coding", list)],
    },
    {
        "table_name": "encounter",
        "column_hierarchy": [("reasoncode", list), ("coding", list)],
    },
    {
        "table_name": "encounter",
        "column_hierarchy": [
            ("hospitalization", dict),
            ("dischargedisposition", dict),
            ("coding", list),
        ],
        "expected": {"dischargedisposition": sql_utils.CODEABLE_CONCEPT},
    },
    # Medication
    {
        "table_name": "medication",
        "column_hierarchy": [("codecodeableconcept", dict), ("coding", list)],
    },
    {
        "table_name": "medication",
        "column_hierarchy": [("medicationcode", dict), ("coding", list)],
    },
    # MedicationRequest
    {
        "table_name": "medicationrequest",
        "column_hierarchy": [("medicationcodeableconcept", dict), ("coding", list)],
    },
    # Observation
    {
        "table_name": "observation",
        "column_hierarchy": [("category", list), ("coding", list)],
    },
    {
        "table_name": "observation",
        "column_hierarchy": [("code", dict), ("coding", list)],
    },
    {
        "table_name": "observation",
        "column_hierarchy": [("component", list), ("code", dict), ("coding", dict)],
        "expected": {"code": sql_utils.CODEABLE_CONCEPT},
    },
    {
        "table_name": "observation",
        "column_hierarchy": [("interpretation", list), ("coding", list)],
    },
    {
        "table_name": "observation",
        "column_hierarchy": [("valuecodeableconcept", dict), ("coding", list)],
    },
    {
        "table_name": "observation",
        "column_hierarchy": [("dataabsentreason", dict), ("coding", list)],
    },
    # Patient
    {
        "table_name": "patient",
        "column_hierarchy": [("maritalstatus", dict), ("coding", list)],
    },
    # Procedure
    {
        "table_name": "procedure",
        "column_hierarchy": [("statusreason", dict), ("coding", list)],
    },
    {
        "table_name": "procedure",
        "column_hierarchy": [("category", dict), ("coding", list)],
    },
    {
        "table_name": "procedure",
        "column_hierarchy": [("code", dict), ("coding", list)],
    },
    {
        "table_name": "procedure",
        "column_hierarchy": [("performer", list), ("function", dict), ("coding", list)],
    },
    {
        "table_name": "procedure",
        "column_hierarchy": [("reasoncode", list), ("coding", list)],
    },
    {
        "table_name": "procedure",
        "column_hierarchy": [("bodysite", list), ("coding", list)],
    },
    {
        "table_name": "procedure",
        "column_hierarchy": [("outcome", dict), ("coding", list)],
    },
    {
        "table_name": "procedure",
        "column_hierarchy": [("complication", list), ("coding", list)],
    },
    {
        "table_name": "procedure",
        "column_hierarchy": [("followup", list), ("coding", list)],
    },
    {
        "table_name": "procedure",
        "column_hierarchy": [("focalDevice", list), ("action", dict), ("coding", list)],
    },
    {
        "table_name": "procedure",
        "column_hierarchy": [("usedcode", list), ("coding", list)],
    },
]



import pathlib

from cumulus_library.template_sql import base_templates

PATH = pathlib.Path(__file__).parent


def get_core_template(
    target_table: str,
    schema: dict[dict[bool]] | None = None,
    config: dict | None = None,
) -> str:
    """Extracts code system details as a standalone table"""
    return base_templates.get_base_template(target_table, path=PATH, schema=schema, config=config)



import cumulus_library
from cumulus_library.studies.core.core_templates import core_templates
from cumulus_library.template_sql import base_templates, sql_utils

expected_table_cols = {
    "condition": {
        "id": [],
        "recordedDate": [],
        "subject": sql_utils.REFERENCE,
        "encounter": sql_utils.REFERENCE,
    }
}


class CoreConditionBuilder(cumulus_library.BaseTableBuilder):
    display_text = "Creating Condition tables..."

    def denormalize_codes(self):
        configs = [
            sql_utils.CodeableConceptConfig(
                source_table="condition",
                column_hierarchy=[("category", list)],
                target_table="core__condition_dn_category",
                # This is an extensible binding, and US Core already suggests three
                # different code systems to pull its recommended four values from.
                # So let's not filter by system here.
            ),
            sql_utils.CodeableConceptConfig(
                source_table="condition",
                column_hierarchy=[("clinicalStatus", dict)],
                target_table="core__condition_dn_clinical_status",
                filter_priority=True,
                code_systems=[
                    # Restrict to just this required binding system
                    "http://terminology.hl7.org/CodeSystem/condition-clinical",
                ],
            ),
            sql_utils.CodeableConceptConfig(
                source_table="condition",
                column_hierarchy=[("code", dict)],
                target_table="core__condition_codable_concepts_display",
                filter_priority=True,
                code_systems=[
                    "http://snomed.info/sct",
                    "http://hl7.org/fhir/sid/icd-10-cm",
                    "http://hl7.org/fhir/sid/icd-9-cm",
                    "http://hl7.org/fhir/sid/icd-9-cm/diagnosis",
                    # EPIC specific systems
                    "urn:oid:1.2.840.114350.1.13.71.2.7.2.728286",
                    "urn:oid:1.2.840.114350.1.13.71.2.7.4.698084.10375",
                    # Spec allowed code of last resort
                    "http://terminology.hl7.org/CodeSystem/data-absent-reason",
                ],
            ),
            sql_utils.CodeableConceptConfig(
                source_table="condition",
                column_hierarchy=[("code", dict)],
                target_table="core__condition_codable_concepts_all",
            ),
            sql_utils.CodeableConceptConfig(
                source_table="condition",
                column_hierarchy=[("verificationStatus", dict)],
                target_table="core__condition_dn_verification_status",
                filter_priority=True,
                code_systems=[
                    # Restrict to just this required binding system
                    "http://terminology.hl7.org/CodeSystem/condition-ver-status",
                ],
            ),
        ]
        self.queries += [
            base_templates.get_codeable_concept_denormalize_query(config) for config in configs
        ]

    def prepare_queries(
        self,
        *args,
        config: cumulus_library.StudyConfig,
        **kwargs,
    ):
        self.denormalize_codes()
        validated_schema = sql_utils.validate_schema(config.db, expected_table_cols)
        self.queries.append(core_templates.get_core_template("condition", validated_schema))



import pathlib
import sys
import tomllib

from cumulus_library import BaseTableBuilder, base_utils, study_manifest
from cumulus_library.builders.valueset import (
    additional_rules_builder,
    rxnorm_valueset_builder,
    static_builder,
    valueset_utils,
)


class ValuesetBuilder(BaseTableBuilder):
    """TableBuilder for creating PSM tables"""

    display_text = "Building valueset tables..."

    def __init__(self, toml_config_path: str, data_path: pathlib.Path, **kwargs):
        """Loads PSM job details from a PSM configuration file"""
        super().__init__()
        # We're stashing the toml path for error reporting later
        self.toml_path = toml_config_path
        self.data_path = data_path
        try:
            with open(self.toml_path, "rb") as file:
                toml_config = tomllib.load(file)

        except OSError:
            sys.exit(f"Valueset configuration not found at {self.toml_path}")
        self.valueset_config = valueset_utils.ValuesetConfig(
            rules_file=toml_config.get("rules_file"),
            keyword_file=toml_config.get("keyword_file"),
            table_prefix=toml_config.get("target_table", ""),
            umls_stewards=toml_config.get("umls_stewards", {}),
            vsac_stewards=toml_config.get("vsac_stewards", {}),
        )

    def prepare_queries(
        self,
        *args,
        config: base_utils.StudyConfig,
        manifest: study_manifest.StudyManifest,
        **kwargs,
    ):
        self.queries = []
        s_builder = static_builder.StaticBuilder()
        s_builder.prepare_queries(
            config=config, manifest=manifest, valueset_config=self.valueset_config
        )
        self.queries += s_builder.queries
        rx_builder = rxnorm_valueset_builder.RxNormValuesetBuilder()
        rx_builder.prepare_queries(
            config=config, manifest=manifest, valueset_config=self.valueset_config
        )
        self.queries += rx_builder.queries
        r_builder = additional_rules_builder.AdditionalRulesBuilder()
        r_builder.prepare_queries(
            config=config, manifest=manifest, valueset_config=self.valueset_config
        )
        self.queries += r_builder.queries



"""Class for communicating with the umls API"""

import os
import pathlib

import requests

from cumulus_library import base_utils, errors

VALID_UMLS_DOWNLOADS = [
    "rxnorm-full-monthly-release",
    "rxnorm-weekly-updates",
    "rxnorm-prescribable-content-monthly-release",
    "rxnorm-prescribable-content-weekly-updates",
    "rxnav-in-a-box",
    "snomed-ct-us-edition",
    "snomed-ct-us-edition-transitive-closure-resources",
    "snomed-ct-international-edition",
    "snomed-ct-core-problem-list-subset",
    "snomed-ct-to-icd-10-cm-mapping-resources",
    "snomed-ct-spanish-edition",
    "umls-metathesaurus-full-subset",
    "umls-metathesaurus-mrconso-file",
    "umls-full-release",
]

VSAC_API_ACTIONS = ["definition", "expansion"]


class UmlsApi:
    def __init__(self, api_key: str | None = None, validator_key: str | None = None):
        """Creates a requests session for future calls, and validates the API key

        :keyword api_key: A UMLS API key (will check for an env var named UMLS_API_KEY
            if None)
        :keyword validator_key: A UMLS API key for the calling application. Will be
            set to the value of api_key if None (which is the current expected
            behavior, since we don't want to be distributing UMLS keys)
        """

        if api_key is None:
            api_key = os.environ.get("UMLS_API_KEY")
            if api_key is None:
                raise errors.ApiError("No UMLS API key provided")
        self.api_key = api_key
        self.validator_key = validator_key or api_key

        auth_payload = {"validatorApiKey": self.validator_key, "apiKey": self.api_key}
        self.session = requests.Session()
        response = self.session.get(
            "https://utslogin.nlm.nih.gov/validateUser", params=auth_payload
        )
        if response.status_code == 401:
            raise errors.ApiError("Invalid UMLS API validator key")
        if response.text != "true":
            raise errors.ApiError("Invalid UMLS API key")
        self.session.auth = requests.auth.HTTPBasicAuth("apikey", api_key)

    def get_vsac_valuesets(
        self,
        url: str | None = None,
        oid: str | None = None,
        action: str = "expansion",
        offset: int = 0,
    ) -> list[dict]:
        """Gets a valueset, and any nested valuesets, from the VSAC API

        :keyword url: an URL to target for a valueset (typically expected)
        :keyword oid: A valuset OID
        :returns: A list, containing the valueset and any referenced
            valuesets.

        Documentation on this API is available at
        https://www.nlm.nih.gov/vsac/support/usingvsac/vsacfhirapi.html

        TODO: do we need to support the v2 API?
        https://www.nlm.nih.gov/vsac/support/usingvsac/vsacsvsapiv2.html
        """
        if action not in VSAC_API_ACTIONS:
            raise errors.ApiError(
                f"get_vsac_valuesets() received unexpected action type {action}. "
                f"Expected values: {','.join(VSAC_API_ACTIONS)}"
            )
        if url is None:
            url = "https://cts.nlm.nih.gov/fhir/res/ValueSet"
        if oid:
            url = f"{url}/{oid}"
        # TODO: Do we need to support the remaining FHIR operators?
        if action == "expansion" and "/$expand" not in url:
            url = url + "/$expand"
        # If we're inspecting url references in a VSAC response, they come back
        # specifying a url that does not align with the actual implemented rest
        # APIs, so we do some massaging
        if "http:" in url:
            url = url.replace("http:", "https:")
        if "/res/" not in url:
            url = url.replace("/fhir/", "/fhir/res/")
        match action:
            case "definition":
                response = self.session.get(url)
            case "expansion":
                response = self.session.get(url + f"?offset={offset}")
        if response.status_code == 404:
            raise errors.ApiError(f"Url not found: {url}")
        all_responses = [response.json()]
        if action == "expansion":
            # Do we need to fetch the next page?
            # Note: the VSAC API hard codes the page size to 1000
            # and ignores the count param in the FHIR ValueSet expand API
            # (https://www.hl7.org/fhir/ValueSet-operation-expand.html),
            # so we just hardcode an expansion of 1000 and hope it
            # doesn't change
            if all_responses[0]["expansion"]["total"] > offset + 1000:
                all_responses += self.get_vsac_valuesets(
                    action=action, url=url, offset=offset + 1000
                )
        if action == "definition":
            included_records = all_responses[0].get("compose", {}).get("include", [])
            for record in included_records:
                if "valueSet" in record:
                    valueset = self.get_vsac_valuesets(action=action, url=record["valueSet"][0])
                    all_responses.append(valueset[0])

        return all_responses

    def get_latest_umls_file_release(self, target: str):
        if target not in VALID_UMLS_DOWNLOADS:
            raise errors.ApiError(
                f"'{target}' is not a valid umls download type.\n\n"
                f"Expected values: {','.join(VALID_UMLS_DOWNLOADS)}"
            )
        release_payload = {"releaseType": target, "current": "true"}
        return self.session.get(
            "https://uts-ws.nlm.nih.gov/releases", params=release_payload
        ).json()[0]

    def download_umls_files(
        self,
        target: str = "umls-metathesaurus-full-subset",
        path: pathlib.Path | None = None,
        unzip: bool = True,
    ):
        """Downloads an available file from the UMLS Download API and unzips it
        target: the UMLS resource to download (default: the MRCONSO.RRF file)
        path: the path on disk to write to

        See https://documentation.uts.nlm.nih.gov/automating-downloads.html for more
        info about the available downloads
        """
        if target not in VALID_UMLS_DOWNLOADS:
            raise errors.ApiError(
                f"'{target}' is not a valid umls download type.\n\n"
                f"Expected values: {','.join(VALID_UMLS_DOWNLOADS)}"
            )
        path = path or pathlib.Path.cwd()
        file_meta = self.get_latest_umls_file_release(target)

        # This particular endpoint requires the API key as a param rather than a
        # basic auth header ¯\_(ツ)_/¯.
        download_payload = {
            "url": file_meta["downloadUrl"],
            "apiKey": self.api_key,
        }
        download_res = requests.get(  # noqa: S113
            "https://uts-ws.nlm.nih.gov/download", params=download_payload, stream=True
        )

        with open(path / file_meta["fileName"], "wb") as f:
            chunks_read = 0
            with base_utils.get_progress_bar() as progress:
                task = progress.add_task(f"Downloading {file_meta['fileName']}", total=None)
                for chunk in download_res.iter_content(chunk_size=1024):
                    f.write(chunk)
                    chunks_read += 1
                    progress.update(
                        task,
                        description=(
                            f"Downloading {file_meta['fileName']}: {chunks_read / 1000} MB"
                        ),
                    )
        if unzip:
            base_utils.unzip_file(path / file_meta["fileName"], path)
            (path / file_meta["fileName"]).unlink()
